__author__ = "Tomasz Rybotycki"

# TODO TR: Consider releasing this file as a separate package.

from typing import List, Union

from numpy import abs, linalg, log2, ndarray, sqrt, pi, exp, asarray, tile, power, diag, dot
from numpy.random import randn


def generate_haar_random_unitary_matrix(d: int) -> ndarray:
    """
        This method generates and returns a random, complex unitary matrix, according
        to the haar measure.

        Note: FBM also gave me a version of that method that generates real matrices,
        not the complex ones.

        :param n: Dimension of the matrix to generate.
        :return: A Haar random unitary matrix.
    """
    random_complex_number = randn(d, d) + 1j * randn(d, d) / sqrt(2)
    q, r = linalg.qr(random_complex_number)
    r = diag(r)
    lamb = diag(r / abs(r))
    q_prime = dot(q, lamb)

    return q_prime.T @ q_prime


def count_total_variation_distance(distribution1: Union[List[float], ndarray],
                                   distribution2: Union[List[float], ndarray]) -> float:
    """
        This method calculates total variation distance between two given distributions.
        :param distribution1: First distribution.
        :param distribution2: Second distribution.
        :return: Total variation distance between two given distributions.
    """

    assert len(distribution1) == len(distribution2), f"Distributions must have equal lengths! Got: {len(distribution1)}" \
                                                     f"and {len(distribution2)}!"
    total_variation_distance = 0

    for i in range(len(distribution1)):
        total_variation_distance += abs(distribution1[i] - distribution2[i])

    return total_variation_distance / 2


def count_distance_between_matrices(matrix1: ndarray, matrix2: ndarray) -> float:
    """
        Calculates distance between two given matrices. This method assumes, that the matrices have proper sizes.
        :param matrix1: First matrix.
        :param matrix2: Second matrix.
        :return: Distance between two given matrices.
    """
    return linalg.norm(matrix1 - matrix2)


def count_tv_distance_error_bound_of_experiment_results(outcomes_number: int, samples_number: int,
                                                        error_probability: float) -> float:
    """
        Calculates the distance bound between the experimental results and the n-sample estimation of these results.

        In case of large outcomes numbers one should consider solutions given here:
        https://math.stackexchange.com/questions/2696344/is-there-a-way-to-find-the-log-of-very-large-numbers

        In the method formally I should compute

            for prime_factor in prime_factors_of_the_large_number:
                error_bound += log2(prime_factor)

        where the large_number =  2 ** outcomes_number - 2.

        However, by simply approximating large_number =  2 ** outcomes_number I can do the same with just

            error_bound += log2(2) * outcomes_number

        or even

            error_bound += outcomes_number

        without wasting a lot of time for calculating the prime factors or the number.

        :param outcomes_number:
        :param samples_number: Number of samples used for estimation.
        :param error_probability: Desired probability of error.
        :return: Bound on the tv distance between the estimate and the experimental results.
    """
    error_bound = -log2(error_probability)
    error_bound += outcomes_number  # APPROXIMATION!

    error_bound /= 2 * samples_number
    return sqrt(error_bound)


def get_prime_factors(number: int) -> List[int]:
    prime_factors = []

    while number % 2 == 0:
        prime_factors.append(2)
        number = number / 2

    for i in range(3, int(sqrt(number)) + 1, 2):
        while number % i == 0:
            prime_factors.append(i)
            number = number / i

    prime_factors.append(number)

    return prime_factors


def compute_minimal_number_of_samples_for_desired_accuracy(outcomes_number: int, error_probability: float,
                                                           expected_distance: float) -> int:
    samples_number = -log2(error_probability)

    samples_number += outcomes_number

    samples_number /= 2 * pow(expected_distance, 2)

    return int(samples_number) + 1


def compute_qft_matrix(n: int) -> ndarray:
    """
        Computes n x n matrix of quantum fourier transform. The formula can be found e.g. on wiki

        https://en.wikipedia.org/wiki/Quantum_Fourier_transform

        :param n: Dimension of the array.
        :return: n x n ndarray of qft.
    """
    if n == 0:
        return asarray([])
    omega = exp(2j * pi / n)

    horizontal_range = tile(range(n), n).reshape(n, n)
    vertical_range = horizontal_range.transpose()
    full_range = horizontal_range * vertical_range
    qft_matrix = power(omega, full_range) / sqrt(n)

    return qft_matrix



__author__ = "Tomasz Rybotycki"

import abc
from dataclasses import dataclass
from typing import List, Iterable

from numpy import ndarray

from ..network_simulation_strategy import network_simulation_strategy


# TODO TR: MO doesn't approve of this class. It should be changed somehow.
@dataclass
class BosonSamplingExperimentConfiguration:
    interferometer_matrix: ndarray  # A matrix describing interferometer.
    initial_state: ndarray
    initial_number_of_particles: int
    number_of_modes: int
    number_of_particles_lost: int
    number_of_particles_left: int
    uniform_transmissivity: float = 0
    network_simulation_strategy: network_simulation_strategy = None
    # TODO TR:  Previously we've used the number of approximated modes instead of the
    #           the hierarchy level. There may be some errors after the changes, that
    #           we should fix.
    hierarchy_level: int = 0  # This is k from papers [1] and [2]


class BSDistributionCalculatorInterface(abc.ABC):
    """ Interface for boson sampling exact distribution calculators """

    @abc.abstractmethod
    def calculate_distribution(self) -> List[float]:
        """ One has to be able to calculate exact distribution with it """
        raise NotImplementedError

    @abc.abstractmethod
    def calculate_probabilities_of_outcomes(self, outcomes: Iterable[Iterable[int]]) -> List[float]:
        """ This method allows one to compute probabilities of only selected outcomes. """
        raise NotImplementedError

    @abc.abstractmethod
    def get_outcomes_in_proper_order(self) -> List[ndarray]:
        """ One also has to know the order of objects that returned probabilities correspond to """
        raise NotImplementedError



__author__ = "Tomasz Rybotycki"

from typing import List

from numpy import arange, ndarray
from numpy.random import choice
from scipy import special

from ..boson_sampling_simulator import BosonSamplingSimulator
from .fixed_loss_simulation_strategy import FixedLossSimulationStrategy
from .simulation_strategy_interface import SimulationStrategyInterface


class UniformLossSimulationStrategy(SimulationStrategyInterface):
    def __init__(self, interferometer_matrix: ndarray,
                 number_of_modes: int, transmissivity: float) \
            -> None:
        self.interferometer_matrix = interferometer_matrix
        self.number_of_modes = number_of_modes
        self.transmissivity = transmissivity

    def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:
        initial_number_of_particles = int(sum(input_state))

        # Using n, eta, l notation from the paper.
        n = initial_number_of_particles
        eta = self.transmissivity

        separable_states_weights = [pow(eta, l) * special.binom(n, l) * pow(1.0 - eta, n - l) for l in range(n + 1)]

        samples = []
        while len(samples) < samples_number:
            number_of_particles_left_in_selected_separable_state = choice(arange(0, n + 1), p=separable_states_weights)

            strategy = FixedLossSimulationStrategy(self.interferometer_matrix,
                                                   number_of_particles_left_in_selected_separable_state,
                                                   self.number_of_modes)

            simulator = BosonSamplingSimulator(strategy)

            samples.append(simulator.get_classical_simulation_results(input_state)[0])

        return samples



__author__ = "Tomasz Rybotycki"

"""
    This class should be used as a base class for all standard BS permanent calculators.
    By standard I mean that the matrix and in(out)put states are stored in a variables.
    It takes care of a lot of boilerplate code.
"""

from typing import Optional

from numpy import ndarray, int64, array, asarray

from ..permanent_calculators.bs_permanent_calculator_interface import BSPermanentCalculatorInterface
import abc


class BSPermanentCalculatorBase(BSPermanentCalculatorInterface, abc.ABC):

    def __init__(self, matrix: ndarray, input_state: Optional[ndarray] = None,
                 output_state: Optional[ndarray] = None) -> None:
        if output_state is None:
            output_state = array([], dtype=int64)
        if input_state is None:
            input_state = array([], dtype=int64)
        self._matrix = matrix
        self._input_state = input_state
        self._output_state = output_state

    @property
    def matrix(self) -> ndarray:
        return self._matrix

    @matrix.setter
    def matrix(self, matrix: ndarray) -> None:
        self._matrix = matrix

    @property
    def input_state(self) -> ndarray:
        return self._input_state

    @input_state.setter
    def input_state(self, input_state: ndarray) -> None:
        self._input_state = asarray(input_state, dtype=int64)

    @property
    def output_state(self) -> ndarray:
        return self._output_state

    @output_state.setter
    def output_state(self, output_state: ndarray) -> None:
        self._output_state = asarray(output_state, dtype=int64)

    def _can_calculation_be_performed(self) -> bool:
        """
            Checks if calculation can be performed. For this to happen sizes of given
            matrix and states have to match.

            :return: Information if the calculation can be performed.
        """
        return self._matrix.shape[0] == self._matrix.shape[1] \
            and len(self._output_state) == len(self._input_state) \
            and len(self._output_state) <= self._matrix.shape[0]



__author__ = "Tomasz Rybotycki"

"""
    This files contains the implementation of Glynn's formula for permanent calculation,
    that uses Gray codes to during the iterations. This method has the same complexity
    as the Ryser formula (O(n2^n), but has been proven to be numerically more stable
    (due to physical limitations of the machines).
"""

try:
    from math import prod
except:
    def prod(values):

        if len(values) == 0:
            return 0

        prod = 1

        for v in values:
            prod *= v

        return prod

from typing import Optional

from numpy import ndarray, complex128, ones

from .bs_permanent_calculator_base import BSPermanentCalculatorBase
from ..boson_sampling_utilities import EffectiveScatteringMatrixCalculator
from ...GuanCodes.src.GrayCode import get_gray_code_update_indices


class GlynnGrayPermanentCalculator(BSPermanentCalculatorBase):

    def __init__(self, matrix: ndarray, input_state: Optional[ndarray] = None,
                 output_state: Optional[ndarray] = None) -> None:
        super().__init__(matrix, input_state, output_state)

    def compute_permanent(self) -> complex128:
        """
            This is the method for computing the permanent. It will work as described
            in [0], meaning implement formula (1) and iterate over deltas in Gray code.
        """

        # Prepare the matrix.
        self.matrix = EffectiveScatteringMatrixCalculator(self.matrix, self.input_state,
                                                          self.output_state).calculate()

        if len(self.matrix) == 0:
            return complex128(1)

        # Initialize the variables.
        delta = ones(len(self.matrix), dtype=int)
        multiplier = 1
        sums = []

        for j in range(len(self.matrix)):
            sums.append(complex128(0))
            for i in range(len(delta)):
                sums[-1] += delta[i] * self.matrix[i][j]

        permanent = multiplier * prod(sums)

        update_indices = get_gray_code_update_indices(len(self.matrix) - 1)

        # Begin the iteration.
        for i in update_indices:
            multiplier = -multiplier
            delta[i] = -delta[i]

            # Sums update
            for j in range(len(sums)):
                sums[j] += 2 * delta[i] * self.matrix[i][j]

            permanent += multiplier * prod(sums)

        return permanent / 2 ** (len(self.matrix) - 1)



__author__ = "Tomasz Rybotycki"

from random import random
from typing import List, Optional

from numpy import conjugate, exp, ndarray, ones, sqrt, zeros, pi
from numpy.random import rand

from .simulation_strategy_interface import SimulationStrategyInterface
from ..network_simulation_strategy.lossy_network_simulation_strategy import LossyNetworkSimulationStrategy
from ..network_simulation_strategy.network_simulation_strategy import NetworkSimulationStrategy


class FixedLossSimulationStrategy(SimulationStrategyInterface):

    def __init__(self, interferometer_matrix: ndarray,
                 number_of_photons_left: int, number_of_observed_modes: int,
                 network_simulation_strategy: Optional[NetworkSimulationStrategy] = None) \
            -> None:
        if network_simulation_strategy is None:
            network_simulation_strategy = LossyNetworkSimulationStrategy(interferometer_matrix)
        self.number_of_photons_left = number_of_photons_left
        self.interferometer_matrix = interferometer_matrix
        self.number_of_observed_modes = number_of_observed_modes
        self._network_simulation_strategy = network_simulation_strategy

    def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:
        """
            Returns an sample from the approximate distribution in fixed losses regime.

            :param samples_number: Number of samples one wants to simulate.
            :param input_state: Usually n-particle Fock state in m modes.
            :return: A sample from the approximation.
        """
        samples = []
        while len(samples) < samples_number:
            phi_0 = self._prepare_initial_state(input_state)
            evolved_state = self._network_simulation_strategy.simulate(input_state=phi_0)
            probabilities = self._calculate_probabilities(evolved_state)
            samples.append(self._calculate_approximation_of_boson_sampling_outcome(probabilities))
        return samples

    def _prepare_initial_state(self, input_state: ndarray) -> ndarray:
        """
            This method is used to prepare psi_0 state (formula 23 from ref. [1]).
            :param input_state: Initial lossy bosonic state.
            :return: Returns the initial state of the formula, which is an equal superposition
            of n photons 'smeared' on the first n modes.
        """
        initial_number_of_photons = int(sum(input_state))
        prepared_state = ones(self.number_of_observed_modes, dtype=float)
        prepared_state[initial_number_of_photons:] = 0
        prepared_state /= sqrt(initial_number_of_photons)  # Note, that numpy version of sqrt is used here!

        return self._randomize_modes_phases(prepared_state)

    @staticmethod
    def _randomize_modes_phases(state_in_modes_basis: ndarray) -> ndarray:
        """
            Randomize the phases of given mode state. Each mode should have different iid random phase.
            :param state_in_modes_basis: A given state in modes basis.
            :return: Given mode state with randomized phases.
        """
        return exp(1j * 2 * pi * rand(len(state_in_modes_basis))) * state_in_modes_basis

    @staticmethod
    def _calculate_probabilities(state: ndarray) -> ndarray:
        return conjugate(state) * state

    def _calculate_approximation_of_boson_sampling_outcome(self, probabilities: ndarray) -> ndarray:
        """
            This method applies evolution to every photon. Note, that evolution of each particle is independent of
            each other.
            :param probabilities:
            :return: A lossy boson state after traversing through interferometer. The state is described in first
            quantization (mode assignment basis).
        """
        output = zeros(self.number_of_observed_modes)
        for photon in range(self.number_of_photons_left):
            x = random()
            i = 0
            prob = probabilities[i]
            while x > prob:
                i += 1
                if len(probabilities) == i:
                    break
                prob += probabilities[i]
            if len(probabilities) != i:
                output[i] += 1
        return output



__author__ = "Tomasz Rybotycki"

from numpy import delete, ndarray, vstack, zeros_like

from .generalized_cliffords_simulation_strategy import GeneralizedCliffordsSimulationStrategy
from .simulation_strategy_interface import SimulationStrategyInterface
from ..boson_sampling_utilities.boson_sampling_utilities import prepare_interferometer_matrix_in_expanded_space
from ..boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_interface import \
    BSPermanentCalculatorInterface


class LossyNetworksGeneralizedCliffordsSimulationStrategy(SimulationStrategyInterface):
    def __init__(self, bs_permanent_calculator: BSPermanentCalculatorInterface) -> None:
        bs_permanent_calculator.matrix = prepare_interferometer_matrix_in_expanded_space(bs_permanent_calculator.matrix)
        self._helper_strategy = GeneralizedCliffordsSimulationStrategy(bs_permanent_calculator)

    def simulate(self, input_state: ndarray, samples_number: int = 1) -> [ndarray]:
        expansion_zeros = zeros_like(input_state)
        expanded_state = vstack([input_state, expansion_zeros])
        expanded_state = expanded_state.reshape(2 * len(input_state), )
        expanded_samples = self._helper_strategy.simulate(expanded_state, samples_number)
        # Trim output state
        samples = []

        for output_state in expanded_samples:
            while len(output_state) > len(input_state):
                output_state = delete(output_state, len(output_state) - 1)
            samples.append(output_state)

        return samples

    def set_new_matrix(self, matrix: ndarray) -> None:
        self._helper_strategy.set_new_matrix(matrix)



__author__ = "Tomasz Rybotycki"

import abc
from typing import List

from numpy import ndarray


class SimulationStrategyInterface(abc.ABC):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, "simulate") and
                callable(subclass.simulate))

    @abc.abstractmethod
    def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:
        """
            Simulate the lossy boson sampling experiment.

            :param input_state: Input state of the simulation.
            :param samples_number: Number of samples one wants to simulate.
            :return:
        """
        raise NotImplementedError



__author__ = "Tomasz Rybotycki"

"""
    This script contains the implementation of the Generalized Cliffords version B
    algorithm. It differs from the original version in two places:
        -   In the k-th step, instead of computing k x k permanents, we compute a set of 
            k-1 x k-1 permanents. From that, we calculate the required k x k permanents
            in O(k) time each.
        -   We permute the columns of the matrix at the beginning. We can then sample
            from an easier distribution and obtain the same results.
"""

from typing import List

from numpy import array, ndarray, int64, zeros_like, complex128
from numpy.random import choice, randint

from .generalized_cliffords_simulation_strategy import \
    GeneralizedCliffordsSimulationStrategy, BSPermanentCalculatorInterface
from ..boson_sampling_utilities.boson_sampling_utilities import \
    modes_state_to_particle_state
from ..boson_sampling_utilities.permanent_calculators.bs_cc_ch_submatrices_permanent_calculator import BSCCCHSubmatricesPermanentCalculator


class GeneralizedCliffordsBSimulationStrategy(GeneralizedCliffordsSimulationStrategy):

    def __init__(self, bs_permanent_calculator: BSPermanentCalculatorInterface) -> None:
        super().__init__(bs_permanent_calculator)
        self._current_input = []
        self._working_input_state = None

    def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:
        """
            Returns sample from linear optics experiments given output state.

            :param input_state: Input state in particle basis.
            :param samples_number: Number of samples to simulate.
            :return: A resultant state after traversing through interferometer.
        """
        self.input_state = input_state
        self.number_of_input_photons = sum(input_state)

        particle_input_state = list(modes_state_to_particle_state(input_state))

        samples = []

        while len(samples) < samples_number:
            self._current_input = zeros_like(input_state)
            self._working_input_state = particle_input_state.copy()
            self._fill_r_sample()
            samples.append(array(self.r_sample, dtype=int64))
        return samples

    def _compute_pmf(self) -> None:

        self.pmf = []

        submatrices_permanents_calculator = BSCCCHSubmatricesPermanentCalculator(
            self._bs_permanent_calculator.matrix, self._current_input, self.r_sample
        )

        submatrices_permanents = submatrices_permanents_calculator.compute_permanents()

        self._bs_permanent_calculator.input_state = self._current_input

        # New particle can come in any new mode
        for m in range(len(self.r_sample)):
            permanent = 0
            for i in range(len(self._current_input)):
                permanent_added = self._current_input[i] * submatrices_permanents[i]
                permanent_added *= self._bs_permanent_calculator.matrix[m][i]
                permanent += permanent_added

            self.pmf.append(abs(permanent) ** 2)

        total = sum(self.pmf)
        self.pmf = [val / total for val in self.pmf]

    def _fill_r_sample(self) -> None:
        self.r_sample = [0 for _ in self.input_state]

        while self.number_of_input_photons > sum(self.r_sample):
            self._update_current_input()
            self._compute_pmf()
            self._sample_from_pmf()

    def _update_current_input(self):
        self._current_input[self._working_input_state.pop(
            randint(0, len(self._working_input_state)))] += 1

    def _sample_from_pmf(self) -> None:
        m = choice(range(len(self.input_state)), p=self.pmf)
        self.r_sample[m] += 1



__author__ = "Tomasz Rybotycki"

"""
    The aim of this class is to produce an implementation of Boson Sampling algorithm
    as in [2], but for uniform losses. This would allow us to apply the losses to the
    k non-approximated particles and, by that, use generalized Cliffords strategy on
    the non-lossy interferometer. We won't need to expand the matrix into 2m x 2m!
"""

from .simulation_strategy_interface import SimulationStrategyInterface
from .generalized_cliffords_b_simulation_strategy import GeneralizedCliffordsBSimulationStrategy, BSPermanentCalculatorInterface
from numpy import ndarray, hstack, zeros_like, complex128, eye, pi, ones, exp, diag, arange
from numpy.random import choice, rand, shuffle
from typing import List
from scipy.special import binom
from ..boson_sampling_utilities.boson_sampling_utilities import generate_lossy_inputs
from multiprocessing import cpu_count
import multiprocessing
from concurrent.futures import ProcessPoolExecutor as Pool
from copy import deepcopy
from ..quantum_computations_utilities import compute_qft_matrix

class LossyStateApproximationSimulationStrategy(SimulationStrategyInterface):
    def __init__(self, bs_permanent_calculator: BSPermanentCalculatorInterface, uniform_transmissivity: float,
                 hierarchy_level: int, threads_number: int = -1):
        # Required for lossy approximated state preparation
        self._approximated_input_state_part_possibilities = None
        self._approximated_input_state_part_possibilities_weights = None
        self._not_approximated_lossy_mixed_state_parts = None
        self._not_approximated_lossy_mixed_state_parts_weights = None

        # Required for general simulation
        self._hierarchy_level = hierarchy_level
        self._uniform_transmissivity = uniform_transmissivity
        self._threads_number = self._get_proper_threads_number(threads_number)
        self._permanent_calculator = bs_permanent_calculator  # Should contain an UNITARY (no losses here!)
        self._qft_matrix = self._get_qft_matrix()

    def _get_proper_threads_number(self, threads_number: int) -> int:
        if threads_number < 1 or threads_number > cpu_count():
            return cpu_count()
        else:
            return threads_number

    def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[
        ndarray]:

        if samples_number < 1:
            return []

        self._prepare_not_approximated_lossy_mixed_state(
            input_state[:self._hierarchy_level]  # Not approximated state part
        )

        self._prepare_approximated_input_state(
            input_state[self._hierarchy_level:]  # Approximated state part
        )

        number_of_samples_for_each_thread = \
            self._compute_number_of_samples_for_each_thread(samples_number)

        # Context is required on Linux systems, as the default (fork) produces undesired results! Spawn is default
        # on osX and Windows and works as expected.
        multiprocessing_context = multiprocessing.get_context("spawn")

        with Pool(mp_context=multiprocessing_context) as p:
            samples_lists = p.map(self._simulate_in_parallel,
                                  number_of_samples_for_each_thread)

        samples = [sample for samples_list in samples_lists for sample in samples_list]

        return samples

    def _prepare_not_approximated_lossy_mixed_state(
            self, not_approximated_input_state_part: ndarray) -> None:
        self._prepare_not_approximated_lossy_mixed_state_parts(not_approximated_input_state_part)
        self._prepare_not_approximated_lossy_mixed_state_parts_weights()

    def _prepare_not_approximated_lossy_mixed_state_parts(
            self, not_approximated_input_state_part: ndarray) -> None:
        self._not_approximated_lossy_mixed_state_parts = []
        for number_of_particles_left in range(sum(not_approximated_input_state_part) + 1):
            self._not_approximated_lossy_mixed_state_parts.extend(generate_lossy_inputs(
                not_approximated_input_state_part, number_of_particles_left))

    def _prepare_not_approximated_lossy_mixed_state_parts_weights(self) -> None:
        # Do note that this method HAS TO be called after lossy mixed state parts are
        # computed.
        possible_weights = self._get_possible_lossy_inputs_weights(
            # Last part is always whole (not approximated) input state part
            self._not_approximated_lossy_mixed_state_parts[-1]
        )
        self._not_approximated_lossy_mixed_state_parts_weights = []

        n = sum(self._not_approximated_lossy_mixed_state_parts[-1])

        for state_part in self._not_approximated_lossy_mixed_state_parts:
            self._not_approximated_lossy_mixed_state_parts_weights.append(
                possible_weights[int(sum(state_part))] / binom(n, sum(state_part))
            )

    def _get_possible_lossy_inputs_weights(
            self, input_state: ndarray) -> List[float]:
        weights = []

        # I'll use the same notation as in [1], for readability.
        n = int(sum(input_state))  # Initial number of particles.
        eta = self._uniform_transmissivity
        for l in range(n + 1):
            # l denotes number of particles left in the state
            weights.append(binom(n, l) * eta ** l * (1 - eta) ** (n - l))

        return weights

    def _prepare_approximated_input_state(
            self, approximated_input_state_part: ndarray) -> None:

        # Assume exact simulation if hierarchy level is not specified.
        if not 0 <= self._hierarchy_level < self._permanent_calculator.matrix.shape[0]:
            self._approximated_input_state_part_possibilities = [[]]
            self._approximated_input_state_part_possibilities_weights = [1]
            return

        self._prepare_approximated_input_state_parts(approximated_input_state_part)
        self._prepare_approximated_input_state_parts_weights()

    def _prepare_approximated_input_state_parts(
            self, approximated_input_state_part: ndarray) -> None:
        self._approximated_input_state_part_possibilities = []
        for number_of_particles_left in range(int(sum(approximated_input_state_part)) + 1):
            state_part_possibility = zeros_like(approximated_input_state_part)
            state_part_possibility[0] = number_of_particles_left
            self._approximated_input_state_part_possibilities.append(
                state_part_possibility
            )

    def _prepare_approximated_input_state_parts_weights(self):
        self._approximated_input_state_part_possibilities_weights = \
            self._get_possible_lossy_inputs_weights(
                # Last part contains all possible particles.
                self._approximated_input_state_part_possibilities[-1]
            )

    def _compute_number_of_samples_for_each_thread(self, samples_number: int) -> List[int]:
        # This method is basically the same as in Nonuniform losses approximation strategy
        samples_per_thread = (samples_number + self._threads_number - (
                    samples_number % self._threads_number))
        samples_per_thread = int(samples_per_thread / self._threads_number)
        number_of_samples_for_each_thread = [samples_per_thread] * self._threads_number

        return number_of_samples_for_each_thread

    def _simulate_in_parallel(self, samples_number: int = 1) -> List[ndarray]:
        """ This method produces given number of samples from lossy approximated
        (separable) state. It's meant to be run in parallel.
        """
        samples = []

        helper_strategy = \
            GeneralizedCliffordsBSimulationStrategy(deepcopy(self._permanent_calculator))

        for _ in range(samples_number):
            lossy_input = self._get_input_state_for_sampling()
            approximate_sampling_matrix = self._get_matrix_for_approximate_sampling()
            helper_strategy.set_new_matrix(approximate_sampling_matrix)
            samples.append(helper_strategy.simulate(lossy_input)[0])

        return samples

    def _get_input_state_for_sampling(self):

        approximated_part = self._approximated_input_state_part_possibilities[choice(
            range(len(self._approximated_input_state_part_possibilities)),
            p=self._approximated_input_state_part_possibilities_weights
        )]

        not_approximated_part = self._not_approximated_lossy_mixed_state_parts[choice(
            range(len(self._not_approximated_lossy_mixed_state_parts)),
            p=self._not_approximated_lossy_mixed_state_parts_weights
        )]
        return hstack([not_approximated_part, approximated_part])

    # Symmetrization fix
    def _permuted_interferometer_matrix(self) -> ndarray:
        permutation = arange(self._permanent_calculator.matrix.shape[0])  # We work with unitary matrices.
        shuffle(permutation)
        return self._permanent_calculator.matrix[:, permutation]

    def _get_matrix_for_approximate_sampling(self) -> ndarray:
        # TODO TR: THIS WILL BE REWRITTEN AFTER MERGING WITH BRUTE-FORCE BRANCH
        random_phases_matrix = self._get_random_phases_matrix()
        return self._permuted_interferometer_matrix() @ random_phases_matrix @ self._qft_matrix

    def _get_qft_matrix(self):
        modes_number = self._permanent_calculator.matrix.shape[0]
        small_qft_matrix = compute_qft_matrix(modes_number - self._hierarchy_level)
        qft_matrix = eye(modes_number, dtype=complex128)

        qft_matrix[self._hierarchy_level:modes_number,
                   self._hierarchy_level:modes_number] = small_qft_matrix

        return qft_matrix

    def _get_random_phases_matrix(self) -> ndarray:
        modes_number = self._permanent_calculator.matrix.shape[0]
        random_phases = ones(modes_number, dtype=complex128)

        random_phases[self._hierarchy_level: modes_number] = \
            exp(1j * 2 * pi * rand(modes_number - self._hierarchy_level))

        return diag(random_phases)





__author__ = "Tomasz Rybotycki"

from numpy import complex128, delete, ndarray, vstack, zeros_like

from .network_simulation_strategy import NetworkSimulationStrategy
from ..boson_sampling_utilities.boson_sampling_utilities import prepare_interferometer_matrix_in_expanded_space


class LossyNetworkSimulationStrategy(NetworkSimulationStrategy):
    def __init__(self, matrix: ndarray) -> None:
        self._matrix = prepare_interferometer_matrix_in_expanded_space(matrix)

    def simulate(self, input_state: ndarray) -> ndarray:
        """
        This method is for simulating lossy network.

        Instead of using NxN matrix, what we need to do is create a 2N x 2N system, and then, depending on singular
        values of passed matrix, we have to transfer some photons into inaccessible, additional modes and then trim
        these additional modes.

        :param input_state: State before parsing through the interferometer. Assume mode occupation basis.
        :return: Lossy output state.
        """
        # Divide by two, because we have 2N x 2N matrix
        input_state = input_state.reshape(self._matrix.shape[0] // 2, 1)
        expansion_zeros = zeros_like(input_state, dtype=complex128)
        expanded_state = vstack([input_state, expansion_zeros])
        evolved_state = self._matrix @ expanded_state
        # Trim the resultant state
        while evolved_state.shape[0] > input_state.shape[0]:
            evolved_state = delete(evolved_state, evolved_state.shape[0] - 1)
        # Reshape to usual space and return.
        evolved_state = evolved_state.flatten()

        return evolved_state



__author__ = "Tomasz Rybotycki"

# TR TODO: Consider making this file a package along with exact distribution calculator.

import itertools
from typing import List, Optional, Sequence

from numpy import array, asarray, block, complex128, diag, eye, int64, ndarray, power, \
    sqrt, transpose, zeros, zeros_like, square, flip, pi, ones, exp
from numpy.linalg import svd
from scipy.special import binom, factorial
from numpy.random import rand

from ..quantum_computations_utilities import compute_qft_matrix


def particle_state_to_modes_state(particle_state: ndarray,
                                  observed_modes_number: int) -> ndarray:
    modes_state = zeros(observed_modes_number)

    # Adding the particle to it's mode.
    for particles_mode in asarray(particle_state, dtype=int64):
        modes_state[particles_mode] += 1

    return modes_state


def modes_state_to_particle_state(mode_state: ndarray) -> ndarray:
    """
        Return given mode-basis state in particle basis.

        :param mode_state: Input state in mode-basis.
        :param particles_number: Number of particles.
        :return: Given mode-basis state in particle basis.
    """

    particles_number = int(sum(mode_state))
    number_of_observed_modes = len(mode_state)
    modes = mode_state.copy()
    particles_state = zeros(particles_number, dtype=int)

    i = k = 0
    while i < number_of_observed_modes:

        if modes[i] > 0:
            modes[i] -= 1
            particles_state[k] = i
            k += 1
        else:
            i += 1

    return particles_state


def generate_possible_outputs(number_of_particles: int, number_of_modes: int,
                              consider_loses: bool = False) \
        -> List[ndarray]:
    if number_of_particles < 0 or number_of_modes < 1:
        return []
    if number_of_particles == 0:
        return [zeros(number_of_modes)]

    outputs = []
    starting_number_of_particles = number_of_particles

    if consider_loses:
        starting_number_of_particles = 0

    for n in range(starting_number_of_particles, number_of_particles + 1):
        outputs.extend(generate_possible_n_particle_outputs(n, number_of_modes))

    return outputs


def generate_possible_n_particle_outputs(number_of_particles: int,
                                         number_of_modes: int) -> List[ndarray]:
    outputs = []

    output = zeros(number_of_modes, dtype=int)
    output[0] = number_of_particles
    outputs.append(output)

    while outputs[-1][number_of_modes - 1] < number_of_particles:

        k = number_of_modes - 1
        while outputs[-1][k - 1] == 0:
            k -= 1

        output = outputs[-1].copy()
        output[k - 1] -= 1
        output[k:] = 0
        output[k] = number_of_particles - sum(output)

        outputs.append(output)

    sorted_outputs = sorted([tuple(output) for output in outputs])

    return [array(output) for output in sorted_outputs]


def generate_lossy_inputs(initial_state: ndarray, number_of_particles_left: int) -> \
List[ndarray]:
    """
        From initial state generate all possible input states after losses application.
        :param initial_state: The state we start with.
        :param number_of_particles_left: Number of particles after losses application.
        :return: A list of lists representing initial states after losses.
    """
    x0 = []
    number_of_modes = len(initial_state)
    initial_number_of_particles = sum(initial_state)
    for i in range(number_of_modes):
        x0.extend([i] * int(initial_state[i]))

    lossy_inputs_list = []
    lossy_inputs_hashes = []

    if sum(initial_state) == 0:
        return [initial_state]

    # Symmetrization.
    for combination in itertools.combinations(list(range(initial_number_of_particles)),
                                              number_of_particles_left):
        lossy_input_in_particle_basis = array([x0[el] for el in combination],
                                              dtype=int64)

        lossy_input = particle_state_to_modes_state(lossy_input_in_particle_basis,
                                                    number_of_modes)

        # Check if calculated lossy input is already in the list. If not, add it.
        lossy_input_hash = hash(tuple(lossy_input))
        if lossy_input_hash not in lossy_inputs_hashes:
            lossy_inputs_list.append(lossy_input)
            lossy_inputs_hashes.append(lossy_input_hash)

    return lossy_inputs_list


def calculate_number_of_possible_n_particle_m_mode_output_states(n: int, m: int) -> int:
    """
        Calculates the number of possible output states with n particles placed around m modes.

        This is basically the same answer as to in how many possible combinations can we put n objects in m bins. It's
        also a dimension of n-particle m-mode bosonic space. Stars-and-bars argument applies here.

        :param n: Number of particles.
        :param m: Number of modes.
        :return: Dimension of n-particle m-mode bosonic space.
    """
    return round(binom(n + m - 1, n))


def calculate_number_of_possible_lossy_n_particle_m_mode_output_states(n: int,
                                                                       m: int) -> int:
    """
        Calculates the number of possible output states with N <= n particles placed around m modes.

        :param n: Number of particles.
        :param m: Number of modes.
        :return: Dimension of n-particle m-mode bosonic space.
    """
    states_number = 0
    for N in range(n + 1):
        states_number += round(binom(N + m - 1, N))
    return states_number


def get_modes_transmissivity_values_from_matrix(lossy_interferometer_matrix: ndarray) -> \
List[float]:
    v_matrix, singular_values, u_matrix = svd(lossy_interferometer_matrix)
    return square(flip(singular_values))


def _calculate_singular_values_matrix_expansion(
        singular_values_vector: ndarray) -> ndarray:
    vector_of_squared_expansions = 1.0 - power(singular_values_vector, 2)
    for i in range(len(vector_of_squared_expansions)):
        if vector_of_squared_expansions[i] < 0:
            vector_of_squared_expansions[i] = 0

    expansion_values = sqrt(vector_of_squared_expansions)

    return diag(expansion_values)


def prepare_interferometer_matrix_in_expanded_space(
        interferometer_matrix: ndarray) -> ndarray:
    v_matrix, singular_values, u_matrix = svd(interferometer_matrix)

    expansions_zeros = zeros_like(v_matrix)
    expansions_ones = eye(len(v_matrix))
    expanded_v = block(
        [[v_matrix, expansions_zeros], [expansions_zeros, expansions_ones]])
    expanded_u = block(
        [[u_matrix, expansions_zeros], [expansions_zeros, expansions_ones]])
    singular_values_matrix_expansion = _calculate_singular_values_matrix_expansion(
        singular_values)
    singular_values_expanded_matrix = block(
        [[diag(singular_values), singular_values_matrix_expansion],
         [singular_values_matrix_expansion, diag(singular_values)]])
    return expanded_v @ singular_values_expanded_matrix @ expanded_u


def compute_state_types(modes_number: int, particles_number: int,
                        losses: bool = False) -> List[List[int]]:
    # Partitions generating code.
    # Taken from https://stackoverflow.com/questions/10035752/elegant-python-code-for-integer-partitioning/10036764
    def partitions(n, I=1):
        yield (n,)
        for i in range(I, n // 2 + 1):
            for p in partitions(n - i, i):
                yield (i,) + p

    all_partitions = list(partitions(particles_number))

    if losses:
        for i in range(particles_number):
            all_partitions += list(partitions(i))

    state_types = []

    for partition in all_partitions:
        if len(partition) > modes_number:
            continue
        # We describe state type by a vector in descending order.
        state_type = sorted(partition, reverse=True)
        state_types.append(state_type)

    for i in range(len(state_types)):
        while len(state_types[i]) < modes_number:
            state_types[i].append(0)

    return state_types


def compute_number_of_state_types(modes_number: int,
                                  particles_number: int,
                                  losses = False) -> int:
    state_types_number = 0

    for k in range(1, modes_number + 1):
        state_types_number += \
            compute_number_of_k_element_integer_n_partitions(k, particles_number)

    if not losses:
        return state_types_number

    for particles_num in range(particles_number):
        for k in range(1, modes_number + 1):
            state_types_number += \
                compute_number_of_k_element_integer_n_partitions(k, particles_num)

    return state_types_number

def compute_number_of_k_element_integer_n_partitions(k: int, n: int) -> int:

    if k == 1:
        return 1

    if k > n or n == 0 or k < 1:
        return 0

    integer_partitions_number = \
        compute_number_of_k_element_integer_n_partitions(k, n - k)
    integer_partitions_number += \
        compute_number_of_k_element_integer_n_partitions(k - 1, n - 1)

    return integer_partitions_number


def compute_maximally_unbalanced_types(modes_number: int, particles_number: int) -> \
List[List[int]]:
    maximally_unbalanced_types = []
    all_types = compute_state_types(particles_number=particles_number,
                                    modes_number=modes_number)

    for state_type in all_types:
        if state_type.count(1) == len(state_type) - 1 or state_type.count(1) == len(
                state_type):
            maximally_unbalanced_types.append(state_type)

    return maximally_unbalanced_types

def compute_number_of_states_of_given_type(state_type: Sequence[int]) -> int:
    modes_number = len(state_type)

    counts = []
    vals = set(state_type)

    for val in vals:
        counts.append(state_type.count(val))

    type_count = factorial(modes_number)

    for count in counts:
        type_count /= factorial(count)

    number_of_states_of_given_type = factorial(modes_number)

    for count in counts:
        number_of_states_of_given_type //= factorial(count)

    return number_of_states_of_given_type


def generate_qft_matrix_for_first_m_modes(m: int, all_modes_number: int) -> ndarray:
    small_qft_matrix = compute_qft_matrix(m)
    qft_matrix = eye(all_modes_number, dtype=complex128)
    qft_matrix[0:m, 0:m] = small_qft_matrix
    return qft_matrix


def generate_random_phases_matrix_for_first_m_modes(m: int, all_modes_number: int) \
        -> ndarray:
    random_phases = ones(all_modes_number, dtype=complex128)  # [1, 1, 1, 1, 1, 1]
    random_phases[0:m] = exp(1j * 2 * pi * rand(m))
    return diag(random_phases)

class EffectiveScatteringMatrixCalculator:
    """
        In many methods of Boson Sampling simulations an effective scattering matrix has to be calculated. Therefore
        I decided to implement an calculator that'd be used if every single one of these methods.
    """

    def __init__(self, matrix: ndarray, input_state: Optional[ndarray] = None,
                 output_state: Optional[ndarray] = None) -> None:
        if output_state is None:
            output_state = array([], dtype=int64)
        if input_state is None:
            input_state = array([], dtype=int64)
        self.__matrix = matrix
        self.__input_state = input_state
        self.__output_state = output_state

    @property
    def matrix(self) -> ndarray:
        return self.__matrix

    @matrix.setter
    def matrix(self, matrix: ndarray) -> None:
        self.__matrix = matrix

    @property
    def input_state(self) -> ndarray:
        return self.__input_state

    @input_state.setter
    def input_state(self, input_state: ndarray) -> None:
        self.__input_state = asarray(input_state, dtype=int64)

    @property
    def output_state(self) -> ndarray:
        return self.__output_state

    @output_state.setter
    def output_state(self, output_state: ndarray) -> None:
        self.__output_state = asarray(output_state, dtype=int64)

    def calculate(self) -> ndarray:

        if sum(self.input_state) == 0 or sum(self.output_state) == 0:
            return []

        transposed_input_matrix = transpose(self.__matrix)
        helper_mtx = []

        for index_of_column_to_insert in range(len(self.__input_state)):
            helper_mtx += [transposed_input_matrix[index_of_column_to_insert]] * \
                          int(self.__input_state[index_of_column_to_insert])

        helper_mtx = transpose(array(helper_mtx, dtype=complex128))

        effective_scattering_matrix = []

        for index_of_row_to_insert in range(len(self.__output_state)):
            effective_scattering_matrix += [helper_mtx[index_of_row_to_insert]] * \
                                           int(self.__output_state[
                                                   index_of_row_to_insert])

        return array(effective_scattering_matrix, dtype=complex128)



__author__ = "Tomasz Rybotycki"

"""
    This file contains implementation of approximate boson sampling strategy subject to non-uniform losses. This can be
    well used to approximate boson sampling experiments with non-balanced network. More details can be found in [2].
"""

import multiprocessing
from concurrent.futures import ProcessPoolExecutor as Pool
from copy import deepcopy
from itertools import repeat
from math import sqrt
from multiprocessing import cpu_count
from typing import List

from numpy import ndarray, diag, ones_like
from numpy.random import choice
from scipy import special

from .lossy_networks_generalized_cliffords_simulation_strategy import \
    BSPermanentCalculatorInterface, \
    LossyNetworksGeneralizedCliffordsSimulationStrategy
from ..boson_sampling_utilities.boson_sampling_utilities import \
    prepare_interferometer_matrix_in_expanded_space, \
    generate_qft_matrix_for_first_m_modes, \
    generate_random_phases_matrix_for_first_m_modes


class NonuniformLossesApproximationStrategy():

    def __init__(self, bs_permanent_calculator: BSPermanentCalculatorInterface,
                 approximated_modes_number: int,
                 modes_transsmisivity: float, threads_number: int = -1) -> None:

        self._approximated_modes_number = self._get_proper_approximated_modes_number(
            bs_permanent_calculator,
            approximated_modes_number)
        self._modes_transmissivity = modes_transsmisivity

        self._initial_matrix = self._prepare_initial_matrix(bs_permanent_calculator)

        self._binom_weights = self._compute_binomial_weights()

        self._threads_number = self._get_proper_threads_number(threads_number)

        self._permanent_calculator = bs_permanent_calculator

    @staticmethod
    def _get_proper_approximated_modes_number(
            bs_permanent_calculator: BSPermanentCalculatorInterface,
            approximated_modes_number: int):
        if approximated_modes_number > bs_permanent_calculator.matrix.shape[0]:
            approximated_modes_number = bs_permanent_calculator.matrix.shape[0]
        if approximated_modes_number < 0:
            approximated_modes_number = 0
        return approximated_modes_number

    def _prepare_initial_matrix(self,
                                bs_permanent_calculator: BSPermanentCalculatorInterface):

        loss_removing_matrix = ones_like(bs_permanent_calculator.matrix[0])
        loss_removing_matrix[:self._approximated_modes_number] = 1.0 / sqrt(
            self._modes_transmissivity) # This here assumes uniform losses
        loss_removing_matrix = diag(loss_removing_matrix)

        initial_matrix = bs_permanent_calculator.matrix @ loss_removing_matrix

        initial_matrix = prepare_interferometer_matrix_in_expanded_space(initial_matrix)

        return initial_matrix

    def _compute_binomial_weights(self):

        eta = self._modes_transmissivity
        k = self._approximated_modes_number

        binom_weights = []

        weight = lambda l: pow(eta, l) * special.binom(k, l) * pow(1.0 - eta, k - l)
        for i in range(k + 1):
            binom_weights.append(weight(i))

        return binom_weights

    def _get_proper_threads_number(self, threads_number: int) -> int:
        if threads_number < 1 or threads_number > cpu_count():
            return cpu_count()
        else:
            return threads_number

    def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:

        if samples_number < 1:
            return []

        # Get samples number per thread
        samples_per_thread = (samples_number + self._threads_number - (
                    samples_number % self._threads_number))
        samples_per_thread = int(samples_per_thread / self._threads_number)
        samples_for_threads = [samples_per_thread] * self._threads_number

        # Context is required on Linux systems, as the default (fork) produces undesired results! Spawn is default
        # on osX and Windows and works as expected.
        multiprocessing_context = multiprocessing.get_context("spawn")

        with Pool(mp_context=multiprocessing_context) as p:
            samples_lists = p.map(self._simulate_in_parallel, repeat(input_state),
                                  samples_for_threads)

        samples = [sample for samples_list in samples_lists for sample in samples_list]

        return samples

    def _simulate_in_parallel(self, input_state: ndarray, samples_number: int = 1):
        samples = []

        helper_strategy = LossyNetworksGeneralizedCliffordsSimulationStrategy(
            deepcopy(self._permanent_calculator))

        for _ in range(samples_number):
            lossy_input = self._compute_lossy_input(input_state)

            # if not array_equal(lossy_input, input_state):
            #    print(f"Got {lossy_input.__str__()}, expected: {input_state.__str__()}") # For k = # modes

            approximate_sampling_matrix = self._get_matrix_for_approximate_sampling()

            # if not array_equal(approximate_sampling_matrix, self._initial_matrix):
            # print(f"Got {approximate_sampling_matrix.__str__()}, expected: {self._initial_matrix.__str__()}")  # For k = # modes

            helper_strategy.set_new_matrix(approximate_sampling_matrix)
            samples.append(helper_strategy.simulate(lossy_input)[0])

        return samples

    def _compute_lossy_input(self, input_state: ndarray) -> ndarray:

        if self._approximated_modes_number < 1:
            return input_state

        lossy_input = deepcopy(input_state)

        binned_input_index = self._approximated_modes_number - 1
        lossy_input[binned_input_index] = choice(
            range(self._approximated_modes_number + 1), p=self._binom_weights)

        return lossy_input

    def _get_matrix_for_approximate_sampling(self) -> ndarray:
        qft_matrix = generate_qft_matrix_for_first_m_modes(
            self._approximated_modes_number,
            self._initial_matrix.shape[0])
        random_phases_matrix = generate_random_phases_matrix_for_first_m_modes(
            self._approximated_modes_number,
            self._initial_matrix.shape[0]
        )

        return self._initial_matrix @ random_phases_matrix @ qft_matrix



__author__ = "Tomasz Rybotycki"

from math import factorial
from random import random
from typing import List

from numpy import array, int64, ndarray, array_equal, isclose
from scipy import special

from .generalized_cliffords_simulation_strategy import GeneralizedCliffordsSimulationStrategy
from ..boson_sampling_utilities.boson_sampling_utilities import generate_possible_outputs
from ..boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_interface import \
    BSPermanentCalculatorInterface


class GeneralizedCliffordsUniformLossesSimulationStrategy(GeneralizedCliffordsSimulationStrategy):
    def __init__(self, bs_permanent_calculator: BSPermanentCalculatorInterface, transmissivity: float = 0):
        self._transmissivity = transmissivity
        self.distribution = []
        self.unweighted_distribution = []
        self._possible_outputs = []
        self._binomial_weights = []
        self.missing_values_in_distribution = False
        super().__init__(bs_permanent_calculator)

    def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:
        """
            Returns sample from linear optics experiments given output state.

            :param input_state: Input state in particle basis.
            :param samples_number: Number of samples to simulate.
            :return: A resultant state after traversing through interferometer.
        """
        self._initialize_simulation(input_state)

        samples = []

        while len(samples) < samples_number:
            self._fill_r_sample()
            samples.append(array(self.r_sample, dtype=int64))
        return samples


    def _initialize_simulation(self, input_state: ndarray) -> None:
        """"
            A method for algorithm initialization before each sampling.
        """

        self.input_state = input_state
        self.number_of_input_photons = sum(input_state)
        self._get_sorted_possible_states()
        self.pmfs = dict()

        distribution_initializer = 0

        if self.missing_values_in_distribution:
            distribution_initializer = -1 # -1 to indicate missing spots

        self._possible_outputs = generate_possible_outputs(sum(input_state), len(input_state), consider_loses=True)
        self.distribution = [distribution_initializer for _ in self._possible_outputs]
        self.unweighted_distribution = [distribution_initializer for _ in self._possible_outputs]

        n = sum(input_state)
        eta = self._transmissivity

        # Do note that index is actually equal to number of particles left!
        self._binomial_weights = \
            [pow(self._transmissivity, left) * special.binom(n, left) * pow(1 - eta, n - left) for left in range(n + 1)]
        self.distribution[0] = self._binomial_weights[0]


    def compute_distribution_up_to_accuracy(self, input_state: ndarray, accuracy: float = 1.0) -> List[float]:
        """
            Returns distribution (up to given accuracy) based on given

            :param input_state: Input state of the experiment.
            :param accuracy: Accuracy up to which distribution will be computed.
            :return:
        """

        self._initialize_simulation(input_state)

        while not isclose(max(accuracy - sum(self.distribution), 0), 0):
            self._fill_r_sample()

        return self.distribution

    def compute_unweighted_distribution_up_to_accuracy(self, input_state: ndarray, accuracy: float = 1.0) -> List[float]:
        """
                    Returns distribution (up to given accuracy) based on given

                    :param input_state: Input state of the experiment.
                    :param accuracy: Accuracy up to which distribution will be computed.
                    :return:
                """

        self._initialize_simulation(input_state)

        while not isclose(max(accuracy - sum(self.unweighted_distribution) / sum(input_state), 0), 0):
            self._fill_r_sample()

        return self.unweighted_distribution

    def _fill_r_sample(self) -> None:
        """
            Fills the r_sample, but it's possible for the photons to be lost.
        """
        self.r_sample = [0 for _ in self._bs_permanent_calculator.matrix]
        self.current_key = tuple(self.r_sample)
        self.current_sample_probability = 1

        for i in range(self.number_of_input_photons):
            if random() >= self._transmissivity:
                continue
            if self.current_key not in self.pmfs:
                self._calculate_new_layer_of_pmfs()
            self._sample_from_latest_pmf()

    def _calculate_new_layer_of_pmfs(self) -> None:
        number_of_particle_to_sample = sum(self.r_sample) + 1
        possible_input_states = self._labeled_states[number_of_particle_to_sample]
        corresponding_k_vectors = [[self.input_state[i] - state[i] for i in range(len(state))]
                                   for state in possible_input_states]

        pmf = []

        weights = self._calculate_weights_from_k_vectors(array(corresponding_k_vectors, dtype=float))
        weights /= sum(weights)
        self.possible_outputs[self.current_key] = self._generate_possible_output_states()

        for output in self.possible_outputs[self.current_key]:
            pmf.append(0)
            for i in range(len(possible_input_states)):
                probability = self._calculate_outputs_probability(possible_input_states[i], output)
                probability *= weights[i]
                pmf[-1] += probability
            for i in range(len(self._possible_outputs)):
                if array_equal(output, self._possible_outputs[i]):
                    self.distribution[i] = pmf[-1]
                    self.distribution[i] *= factorial(sum(output))
                    for val in output:
                        self.distribution[i] /= factorial(val)
                    self.unweighted_distribution[i] = self.distribution[i]
                    self.distribution[i] *= self._binomial_weights[sum(output)]

        self.pmfs[self.current_key] = pmf



__author__ = "Tomasz Rybotycki"

from numpy import dot, ndarray

from .network_simulation_strategy import NetworkSimulationStrategy


class LosslessNetworkSimulationStrategy(NetworkSimulationStrategy):
    def __init__(self, matrix: ndarray) -> None:
        self._matrix = matrix

    def simulate(self, input_state: ndarray) -> ndarray:
        return dot(self._matrix, input_state)



__author__ = "Tomasz Rybotycki"

"""
    This file holds the implementation of Ryser method of permanent calculation, as
    presented in [Cliffords2020]. I've modified it slightly, so that the permanent can
    be used for computing permanents for bunched and not-continuous 
    (like [1, 1, 0, 1, 0]) inputs.
"""

from typing import List, Optional

from numpy import complex128, ndarray
from scipy.special import binom

from ..permanent_calculators.bs_permanent_calculator_base import \
    BSPermanentCalculatorBase


class RyserPermanentCalculator(BSPermanentCalculatorBase):
    """
        This class is designed to calculate permanent of effective scattering matrix of
        a boson sampling instance. Note, that it can be used to calculate permanent of
        a given matrix. All that is required that input and output states are correct
        representations of Fock states with proper dimensions.
    """

    def __init__(self, matrix: ndarray, input_state: Optional[ndarray] = None,
                 output_state: Optional[ndarray] = None) -> None:
        super().__init__(matrix, input_state, output_state)

    def compute_permanent(self) -> complex128:
        """
            This is the main method of the calculator. Assuming that input state,
            output state and the matrix are defined correctly (that is we've got m x m
            matrix, and vectors of with length m) this calculates the permanent of an
            effective scattering matrix related to probability of obtaining output state
            from given input state.

            :return: Permanent of effective scattering matrix.
        """
        if not self._can_calculation_be_performed():
            raise AttributeError

        r_vectors = self._calculate_r_vectors()

        permanent = complex128(0)
        for r_vector in r_vectors:
            permanent += self._compute_permanent_addend(r_vector)
        permanent *= pow(-1, sum(self._input_state))
        return permanent

    def _calculate_r_vectors(self,
                             input_vector: Optional[ndarray] = None) -> List[ndarray]:
        """
            This method is used to calculate all the r vectors that appear in the
            Ryser's formula during permanent calculation. By r vectors I denote vectors
            in form [r_1, r_2, ..., r_m]. This method basically takes care of the sum of
            sum ... of sum at the beginning of the formula.

            :return: List of r vectors.
        """
        if input_vector is None:
            input_vector = []
        r_vectors = []
        for i in range(int(self._output_state[len(input_vector)]) + 1):
            input_state = input_vector.copy()
            input_state.append(i)

            if len(input_state) == len(self._output_state):
                r_vectors.append(input_state)
            else:
                r_vectors.extend(self._calculate_r_vectors(input_state))

        return r_vectors

    def _compute_permanent_addend(self, r_vector: ndarray) -> complex128:

        addend = pow(-1, sum(r_vector))

        # Binomials calculation
        for i in range(len(r_vector)):
            addend *= binom(self._output_state[i], r_vector[i])

        # Product calculation
        product = 1

        considered_columns_indices = []
        for mode_index in range(len(self._input_state)):
            if self._input_state[mode_index] != 0:
                considered_columns_indices.append(
                    mode_index)  # Consider non-standard inputs

        for j in considered_columns_indices:
            product_part = 0

            for nu in range(len(self._input_state)):
                product_part += r_vector[nu] * self._matrix[nu][j]

            product *= pow(product_part, self._input_state[
                j])  # Take into account bunching in the input
        addend *= product
        return addend



__author__ = "Tomasz Rybotycki"

"""
    This file hold a factory for BS permanent calculators.
"""

import enum
from typing import Optional

from numpy import ndarray

from .bs_permanent_calculator_interface import BSPermanentCalculatorInterface
from .chin_huh_permanent_calculator import ChinHuhPermanentCalculator
from .classic_permanent_calculator import ClassicPermanentCalculator
from .parallel_chin_huh_permanent_calculator import ParallelChinHuhPermanentCalculator
from .ryser_guan_permanent_calculator import RyserGuanPermanentCalculator
from .ryser_permanent_calculator import RyserPermanentCalculator


class PermanentCalculatorType(enum.IntEnum):
    CLASSIC = enum.auto()
    CHIN_HUH = enum.auto()
    PARALLEL_CHIN_HUH = enum.auto()
    RYSER = enum.auto()
    RYSER_GUAN = enum.auto()


class BSPermanentCalculatorFactory:
    def __init__(self, matrix: Optional[ndarray], input_state: Optional[ndarray],
                 output_state: Optional[ndarray],
                 calculator_type: PermanentCalculatorType = PermanentCalculatorType.PARALLEL_CHIN_HUH):
        self._matrix = matrix
        self._input_state = input_state
        self._output_state = output_state

        self._calculator_type = calculator_type
        self._calculator_mapping = {
            PermanentCalculatorType.CLASSIC: self._generate_classic_permanent_calculator,
            PermanentCalculatorType.CHIN_HUH: self._generate_chin_huh_permanent_calculator,
            PermanentCalculatorType.PARALLEL_CHIN_HUH: self._generate_parallel_chin_huh_permanent_calculator,
            PermanentCalculatorType.RYSER: self._generate_ryser_permanent_calculator,
            PermanentCalculatorType.RYSER_GUAN: self._generate_ryser_guan_permanent_calculator
        }

    @property
    def matrix(self) -> ndarray:
        return self._matrix

    @matrix.setter
    def matrix(self, new_matrix: ndarray) -> None:
        self._matrix = new_matrix

    @property
    def input_state(self) -> ndarray:
        return self._input_state

    @input_state.setter
    def input_state(self, new_input_state: ndarray) -> None:
        self._input_state = new_input_state

    @property
    def output_state(self) -> ndarray:
        return self._output_state

    @output_state.setter
    def output_state(self, new_output_state: ndarray) -> None:
        self._output_state = new_output_state

    def generate_calculator(self) -> BSPermanentCalculatorInterface:
        """
            Generates the permanent as specified by the provided calculator type.
        """
        handler = self._calculator_mapping.get(self._calculator_type,
                                               self._generate_parallel_chin_huh_permanent_calculator)
        return handler()

    def _generate_classic_permanent_calculator(self) -> ClassicPermanentCalculator:
        return ClassicPermanentCalculator(matrix=self.matrix,
                                          input_state=self.input_state,
                                          output_state=self.output_state)

    def _generate_chin_huh_permanent_calculator(self) -> ChinHuhPermanentCalculator:
        return ChinHuhPermanentCalculator(matrix=self.matrix,
                                          input_state=self.input_state,
                                          output_state=self.output_state)

    def _generate_parallel_chin_huh_permanent_calculator(
            self) -> ParallelChinHuhPermanentCalculator:
        return ParallelChinHuhPermanentCalculator(matrix=self.matrix,
                                                  input_state=self.input_state,
                                                  output_state=self.output_state)

    def _generate_ryser_permanent_calculator(self) -> RyserPermanentCalculator:
        return RyserPermanentCalculator(matrix=self.matrix,
                                        input_state=self.input_state,
                                        output_state=self.output_state)

    def _generate_ryser_guan_permanent_calculator(self) -> RyserGuanPermanentCalculator:
        return RyserGuanPermanentCalculator(matrix=self.matrix,
                                            input_state=self.input_state,
                                            output_state=self.output_state)



__author__ = "Tomasz Rybotycki"

"""
    This class is used to calculate permanent of effective scattering matrix. It first generates the matrix, and then
    calculates the permanent via standard means.
"""

from typing import List, Optional

from numpy import complex128, ndarray

from ..boson_sampling_utilities import EffectiveScatteringMatrixCalculator
from ...boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_base import BSPermanentCalculatorBase


class ClassicPermanentCalculator(BSPermanentCalculatorBase):

    def __init__(self, matrix: ndarray, input_state: Optional[ndarray] = None,
                 output_state: Optional[ndarray] = None) -> None:
        super().__init__(matrix, input_state, output_state)

    def compute_permanent(self) -> complex128:

        if sum(self.input_state) == 0:
            if sum(self.output_state) == 0:
                return complex128(1 + 0j)
            else:
                return complex128(0)

        scattering_matrix_calculator = \
            EffectiveScatteringMatrixCalculator(self._matrix, self._input_state, self._output_state)
        scattering_matrix = scattering_matrix_calculator.calculate()
        return self._compute_permanent_recursively(scattering_matrix)

    def _compute_permanent_recursively(self, matrix: ndarray) -> complex128:
        """
        Returns the permanent of the matrix.
        """
        return self._permanent_recursive_part(matrix, column=0, selected=[], prod=complex128(1))

    def _permanent_recursive_part(self, mtx: ndarray, column: int, selected: List[int], prod: complex128) -> complex128:
        """
        Row expansion for the permanent of matrix mtx.
        The counter column is the current column,
        selected is a list of indices of selected rows,
        and prod accumulates the current product.
        """
        if column == mtx.shape[1]:
            return prod

        result = complex128(0 + 0j)
        for row in range(mtx.shape[0]):
            if row not in selected:
                result += self._permanent_recursive_part(mtx, column + 1, selected + [row], prod * mtx[row, column])
        return result



__author__ = "Tomasz Rybotycki"

import abc

from numpy import ndarray


class NetworkSimulationStrategy(abc.ABC):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, "simulate") and
                callable(subclass.simulate))

    @abc.abstractmethod
    def simulate(self, input_state: ndarray) -> ndarray:
        raise NotImplementedError



__author__ = "Tomasz Rybotycki"

from collections import defaultdict
from copy import copy
from math import factorial
from typing import List
from scipy.special import binom

from numpy import array, delete, float64, insert, int64, ndarray
from numpy.random import random

from .simulation_strategy_interface import SimulationStrategyInterface
from ..boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_interface import \
    BSPermanentCalculatorInterface


class GeneralizedCliffordsSimulationStrategy(SimulationStrategyInterface):
    def __init__(self, bs_permanent_calculator: BSPermanentCalculatorInterface) -> None:
        self.r_sample = []
        self.number_of_input_photons = 0
        self.pmfs = dict()  # Probability mass functions calculated along the way. Keys should be current r as tuples.
        self._bs_permanent_calculator = bs_permanent_calculator
        self.input_state = array([], dtype=int64)
        self._labeled_states = defaultdict(list)
        self.possible_outputs = dict()
        self.current_key = tuple(self.r_sample)
        self.current_sample_probability = 1

    def set_new_matrix(self, new_matrix: ndarray) -> None:
        self._bs_permanent_calculator.matrix = new_matrix

    def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:
        """
            Returns sample from linear optics experiments given output state.

            :param input_state: Input state in particle basis.
            :param samples_number: Number of samples to simulate.
            :return: A resultant state after traversing through interferometer.
        """
        self.input_state = input_state
        self.number_of_input_photons = sum(input_state)
        self._get_sorted_possible_states()
        self.pmfs = dict()

        samples = []

        while len(samples) < samples_number:
            self._fill_r_sample()
            samples.append(array(self.r_sample, dtype=int64))
        return samples

    def _get_sorted_possible_states(self) -> None:
        """
            Calculate and sort all the substates of the input. They will later be used to calculate output
            probabilities.

            :return: Dict of all the possible substates of input (including 0 and the state), labeled with number of
            particles in this state.
        """
        # Calculating all possible substates of the input
        possible_input_states = self._calculate_all_input_substates(self.input_state.copy())

        # Labeling them into dict where keys are being number of particles in the state.
        self._labeled_states = defaultdict(list)

        for state in possible_input_states:
            states_particles_number = sum(state)
            self._labeled_states[states_particles_number].append(state)

    def _calculate_all_input_substates(self, state_part_left: ndarray) -> List[ndarray]:
        """
        Calculates substates of the input in recursive manner.

        :param state_part_left: State with reduced modes number.
        :return: All the substates for starting number of modes.
        """
        if len(state_part_left) < 1:
            return [array([], dtype=int64)]

        n = state_part_left[0]
        state_part_left = delete(state_part_left, 0)

        smaller_substates = self._calculate_all_input_substates(state_part_left.copy())

        substates = []
        for i in range(int(n + 1)):
            for substate in smaller_substates:
                new_substate = substate.copy()
                new_substate = insert(new_substate, 0, i)
                substates.append(new_substate)
        return substates

    def _fill_r_sample(self) -> None:
        self.r_sample = [0 for _ in self.input_state]
        self.current_key = tuple(self.r_sample)
        self.current_sample_probability = 1

        while self.number_of_input_photons > sum(self.r_sample):
            if self.current_key not in self.pmfs:
                self._calculate_new_layer_of_pmfs()
            self._sample_from_latest_pmf()

    def _calculate_new_layer_of_pmfs(self) -> None:

        number_of_particle_to_sample = sum(self.r_sample) + 1

        possible_input_states = self._labeled_states[number_of_particle_to_sample]

        corresponding_k_vectors = [[self.input_state[i] - state[i] for i in range(len(state))]
                                   for state in possible_input_states]

        weights = self._calculate_weights_from_k_vectors(array(corresponding_k_vectors, dtype=float))

        weights /= sum(weights)
        self.possible_outputs[self.current_key] = self._generate_possible_output_states()

        pmf = []

        for output in self.possible_outputs[self.current_key]:
            pmf.append(0)
            for i in range(len(possible_input_states)):
                probability = self._calculate_outputs_probability(possible_input_states[i], output)
                probability *= weights[i]
                pmf[-1] += probability

        self.pmfs[self.current_key] = pmf

    def _calculate_weights_from_k_vectors(self, corresponding_k_vectors: ndarray) -> ndarray:
        return array([self._calculate_weights(vector)
                      for vector in corresponding_k_vectors], dtype=float64)

    def _calculate_weights(self, k_vector):
        l = sum(k_vector)
        n = sum(self.input_state)

        weight = factorial(l) * factorial(n - l) / factorial(n)

        for m in range(len(self.input_state)):
            weight *= binom(self.input_state[m], k_vector[m])

        return weight

    def _generate_possible_output_states(self) -> List[ndarray]:
        possible_output_states = []

        for i in range(len(self.r_sample)):
            new_possible_output = copy(self.r_sample)
            new_possible_output[i] += 1
            possible_output_states.append(array(new_possible_output, dtype=int64))

        return possible_output_states

    def _calculate_outputs_probability(self, input_state: ndarray, output_state: ndarray) -> float:
        self._bs_permanent_calculator.input_state = input_state
        self._bs_permanent_calculator.output_state = output_state
        probability = abs(self._bs_permanent_calculator.compute_permanent()) ** 2

        for mode_occupation_number in input_state:
            probability /= factorial(mode_occupation_number)

        probability /= factorial(sum(input_state))

        return probability

    def _sample_from_latest_pmf(self) -> None:

        sample_index = 0
        random_value = random() * sum(self.pmfs[self.current_key])  # PMFs are not normalized.
        current_probability = 0
        for probability in self.pmfs[self.current_key]:
            current_probability += probability
            if current_probability > random_value:
                break
            sample_index += 1

        self.current_sample_probability = self.pmfs[self.current_key][sample_index]
        self.r_sample = self.possible_outputs[self.current_key][sample_index]
        self.current_key = tuple(self.r_sample)



__author__ = "Tomasz Rybotycki"

import math
from copy import deepcopy
from typing import List, Iterable

from numpy import ndarray
from scipy import special

from ..boson_sampling_utilities.boson_sampling_utilities import generate_lossy_inputs, generate_possible_outputs
from ..boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_interface import \
    BSPermanentCalculatorInterface
from ..distribution_calculators.bs_distribution_calculator_interface import BosonSamplingExperimentConfiguration, \
    BSDistributionCalculatorInterface


class BSDistributionCalculatorWithFixedLosses(BSDistributionCalculatorInterface):
    def __init__(self, configuration: BosonSamplingExperimentConfiguration,
                 permanent_calculator: BSPermanentCalculatorInterface) -> None:
        self.configuration = deepcopy(configuration)
        self._permanent_calculator = permanent_calculator

    def get_outcomes_in_proper_order(self) -> List[ndarray]:
        return generate_possible_outputs(self.configuration.number_of_particles_left,
                                         self.configuration.number_of_modes)

    def calculate_distribution(self) -> List[float]:
        """
        This method will be used to calculate the exact distribution of lossy boson sampling experiment.
        The results will be returned as a table of probabilities of obtaining the outcome at i-th index.
        :return: List of probabilities of outcomes.
        """

        possible_outcomes = self.get_outcomes_in_proper_order()

        return self.calculate_probabilities_of_outcomes(possible_outcomes)

    def calculate_probabilities_of_outcomes(self, outcomes: Iterable[Iterable[int]] ) -> List[float]:
        outcomes_probabilities = [self.__calculate_probability_of_outcome(outcome) for
                                  outcome in outcomes]

        return outcomes_probabilities

    def __calculate_probability_of_outcome(self, outcome: ndarray) -> float:
        """
        Given input state and losses as described in the BS experiment configuration
        compute the probability of obtaining the outcome given as a parameter, after
        traversing a lossy linear-optical channel.
        :param outcome: A Fock state for which the probability will be computed.
        :return: Probability of obtaining given state in current experiment config.
        """
        outcome_probability = self.__calculate_probability_of_outcome_state_for_indistinguishable_photons(outcome)

        # Different states in particles-basis may give the same outcome state.
        outcome_probability *= math.factorial(self.configuration.number_of_particles_left)
        for i in range(self.configuration.number_of_modes):
            outcome_probability /= math.factorial(outcome[i])

        return outcome_probability

    def __calculate_probability_of_outcome_state_for_indistinguishable_photons(self, outcome_state: ndarray) -> float:
        probability_of_outcome = 0

        # Symmetrize the input.
        lossy_inputs_list = generate_lossy_inputs(self.configuration.initial_state,
                                                  self.configuration.number_of_particles_left)

        for lossy_input in lossy_inputs_list:
            self._permanent_calculator.matrix = self.configuration.interferometer_matrix
            self._permanent_calculator.input_state = lossy_input
            self._permanent_calculator.output_state = outcome_state
            subprobability = abs(self._permanent_calculator.compute_permanent()) ** 2
            for mode_occupation_number in lossy_input:
                subprobability /= math.factorial(mode_occupation_number)

            probability_of_outcome += subprobability

        # Normalization (read Brod & Oszmaniec 2019).
        probability_of_outcome /= math.factorial(self.configuration.number_of_particles_left)
        probability_of_outcome /= special.binom(self.configuration.initial_number_of_particles,
                                                self.configuration.number_of_particles_left)

        return probability_of_outcome



__author__ = "Tomasz Rybotycki"

"""
    This script contains the implementation of distribution calculator for special
    separable states described in [1] i [2]. For a given matrix U we would expect
    that the distribution returned would be _good_ approximation of U.
"""

from typing import Iterable, List

from numpy import ndarray, ones_like, sqrt, diag, zeros, hstack

from .bs_distribution_calculator_interface import BSDistributionCalculatorInterface, \
    BosonSamplingExperimentConfiguration
from ..boson_sampling_utilities.boson_sampling_utilities import \
    generate_possible_outputs, generate_qft_matrix_for_first_m_modes, \
    generate_random_phases_matrix_for_first_m_modes, \
    prepare_interferometer_matrix_in_expanded_space
from ..boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_interface import \
    BSPermanentCalculatorInterface

from .bs_distribution_calculator_with_fixed_losses import BSDistributionCalculatorWithFixedLosses


class BSApproximatedLossyDistributionCalculator(BSDistributionCalculatorInterface):

    def __init__(self, configuration: BosonSamplingExperimentConfiguration,
                 permanent_calculator: BSPermanentCalculatorInterface) -> None:
        self._configuration = configuration
        self._permanent_calculator = permanent_calculator
        self._initial_matrix = self._prepare_initial_matrix()
        self._initial_state = self._prepare_initial_state()
        # How many matrices should I average over to get the distribution?
        # TODO TR: How many of these should I actually put here?
        self._matrices_number = 100


    def _prepare_initial_matrix(self):
        # TODO TR: This is basically the same as in nonuniform_losses_approximation_strategy.
        # I should do something about that.
        loss_removing_matrix = ones_like(self._configuration.interferometer_matrix[0])
        loss_removing_matrix[:self._approximated_modes_number] = 1.0 / sqrt(
            self._modes_transmissivity)  # This here assumes uniform losses
        loss_removing_matrix = diag(loss_removing_matrix)

        initial_matrix = self._configuration.interferometer_matrix @ loss_removing_matrix

        initial_matrix = prepare_interferometer_matrix_in_expanded_space(initial_matrix)

        return initial_matrix

    def _prepare_initial_state(self):
        """
            Initial state of approximate simulation (before applying qft and random
            phases) should have all the particles from approximated modes (we assume
            that we approximate first m modes) in the last approximated mode.

            Something like:
            [0, 0, ..., 0, k, NON-APPROXIMATED REST OF THE STATE]

            For simplicity I also extend it to 2m.
        """

        # Extend the state to 2m
        initial_state = zeros(2 * self._configuration.number_of_modes, dtype=int)
        initial_state[:self._configuration.number_of_modes] = self._configuration.initial_state

        # Bunch approximated modes
        number_of_particles_to_put_in_the_last_approximated_mode = 0
        for i in range(self._configuration.approximated_modes_number - 1):
            number_of_particles_to_put_in_the_last_approximated_mode += initial_state[i]
            initial_state[i] = 0

        initial_state[self._configuration.approximated_modes_number - 1] = \
            number_of_particles_to_put_in_the_last_approximated_mode

        return initial_state

    def calculate_distribution(self) -> List[float]:
        """
            Computes whole distribution basing on configuration.
        """
        possible_outcomes = generate_possible_outputs(
            self.configuration.number_of_particles_left,
            self.configuration.number_of_modes)
        return self.calculate_probabilities_of_outcomes(possible_outcomes)

    def calculate_probabilities_of_outcomes(self, outcomes: Iterable[Iterable[int]]) -> \
            List[float]:
        probabilities = []

        # TODO TR: We could probably make this parallel.
        for _ in range(self._matrices_number):
            effective_matrix = self._generate_effective_matrix()
            for outcome in outcomes:
                probabilities.append(self._calculate_probability_of_outcome(outcome, effective_matrix))

        return probabilities

    def _generate_effective_matrix(self):
        effective_matrix = generate_qft_matrix_for_first_m_modes(
            self._configuration.approximated_modes_number,
            self._configuration.number_of_modes * 2)
        effective_matrix = effective_matrix @ generate_random_phases_matrix_for_first_m_modes(
            self._configuration.approximated_modes_number,
            self._configuration.number_of_modes * 2)
        effective_matrix = effective_matrix @ self._initial_matrix
        return effective_matrix

    def _calculate_probability_of_outcome(self, outcome: Iterable[int],
                                          matrix: ndarray) -> float:
        # First thing to note is that I can calculate probabilities of 2m-mode states,
        # but what I need is m-mode states' probabilities.

        # As a matter of fact I can compute 1m probability from 2m probability, by
        # summing probabilities of these states that have outcome state on first m modes
        # (first m, because we prepare the effective matrix in such a way).
        considered_outputs = self._get_2m_outcomes_corresponding_to_the_outcome(outcome)

        # I want to compute lossless distribution (or even probabilities of the lossless
        # outputs), so I may use one of the distribution calculators that I've got.
        subproblem_configuration = BosonSamplingExperimentConfiguration(
            interferometer_matrix=matrix,
            initial_state=self._initial_state,
            initial_number_of_particles=self._configuration.initial_number_of_particles,
            number_of_modes=2*self._configuration.number_of_modes,
            number_of_particles_lost=0,
            number_of_particles_left=self._configuration.initial_number_of_particles
        )
        subdistribution_calculator = BSDistributionCalculatorWithFixedLosses(
            permanent_calculator=self._permanent_calculator,
            configuration=subproblem_configuration
        )
        probabilities_of_outcomes = subdistribution_calculator.calculate_probabilities_of_outcomes(
            outcomes=considered_outputs
        )
        return sum(probabilities_of_outcomes)

    def _get_2m_outcomes_corresponding_to_the_outcome(self, outcome: Iterable[int]) -> List[Iterable[int]]:
        """
            This method returns 2m outcomes for given particle number, such that
            in the first m modes we've got exactly output state. Assuming that there
            are n particles total and k particles in the given outcome, we compute
            all m-mode (n-k)-particles states and just hstack them with output.
        """
        considered_outcomes = []
        possible_m_mode_outputs_with_less_particles = \
            generate_possible_outputs(
                number_of_particles=self._configuration.initial_number_of_particles - sum(outcome),
                number_of_modes=self._configuration.number_of_modes
            )

        for possible_output in possible_m_mode_outputs_with_less_particles:
            considered_outcomes.append(hstack([outcome, possible_output]))

        return considered_outcomes


    def get_outcomes_in_proper_order(self) -> List[ndarray]:
        """
            Returns states in the same order that distribution probabilities were
            calculated in.
        """
        return generate_possible_outputs(self._configuration.number_of_particles_left,
                                         self._configuration.number_of_modes)



__author__ = "Tomasz Rybotycki"

"""
    This file holds the implementation of Ryser method of permanent calculation, as
    presented in [Cliffords2020] with using Guan codes to speed it up. I modified it
    slightly, so that the permanent can be used for computing permanents for bunched and
    not-continuous (like [1, 1, 0, 1, 0]) inputs.
"""

import operator
from functools import reduce
from typing import Optional

from numpy import complex128, ndarray, nonzero, zeros, ones

from ..permanent_calculators.bs_permanent_calculator_base import \
    BSPermanentCalculatorBase


class RyserGuanPermanentCalculator(BSPermanentCalculatorBase):
    """
        This class is designed to calculate permanent of effective scattering matrix of
        a boson sampling instance. Note, that it can be used to calculate permanent of
        a given matrix. All that is required that input and output states are correct
        representations of Fock states with proper dimensions.
    """

    def __init__(self, matrix: ndarray, input_state: Optional[ndarray] = None,
                 output_state: Optional[ndarray] = None) -> None:
        super().__init__(matrix, input_state, output_state)

    def compute_permanent(self) -> complex128:
        """
            This is the main method of the calculator. Assuming that input state, output
            state and the matrix are defined correctly (that is we've got m x m matrix,
            and vectors of with length m) this calculates the permanent of an effective
            scattering matrix related to probability of obtaining output state from
            a given input state.

            :return: Permanent of effective scattering matrix.
        """
        if not self._can_calculation_be_performed():
            raise AttributeError

        r_vector = zeros(len(self._output_state), dtype=int)  # g
        code_update_information = ones(len(self._output_state), dtype=int)  # u
        position_limits = list(self._output_state)  # n

        permanent = complex128(0)
        sums = dict()

        considered_columns_indices = nonzero(self.input_state)[0]

        multiplier = 1
        binomials_product = 1

        # Initialization (0-th step).
        for i in considered_columns_indices:
            sums[i] = 0
            for j in range(len(self._input_state)):
                sums[i] += self._input_state[j] * self._matrix[i][j]

        permanent += multiplier * binomials_product * \
                     reduce(operator.mul, [pow(sums[i], self._input_state[i]) for i in considered_columns_indices], 1)

        while r_vector[-1] <= position_limits[-1]:

            # UPDATE R VECTOR
            index_to_update = 0  # i
            updated_value_at_index = r_vector[0] + code_update_information[0]  # k
            while updated_value_at_index > position_limits[
                index_to_update] or updated_value_at_index < 0:
                code_update_information[index_to_update] = -code_update_information[
                    index_to_update]
                index_to_update += 1

                if index_to_update == len(r_vector):
                    permanent *= pow(-1, sum(self._input_state))
                    return permanent

                updated_value_at_index = r_vector[index_to_update] + \
                                         code_update_information[index_to_update]

            last_value_at_index = r_vector[index_to_update]
            r_vector[index_to_update] = updated_value_at_index
            # END UPDATE

            # START PERMANENT UPDATE
            multiplier = -multiplier

            # Sums update
            for j in sums:
                sums[j] += (r_vector[index_to_update] - last_value_at_index) * \
                           self.matrix[index_to_update][j]

            # Binoms update
            if r_vector[index_to_update] > last_value_at_index:
                binomials_product *= (self._output_state[index_to_update] -
                                      last_value_at_index) / r_vector[index_to_update]
            else:
                binomials_product *= last_value_at_index / (
                            self._output_state[index_to_update] - r_vector[index_to_update])

            permanent += multiplier * binomials_product * reduce(operator.mul,
                                                                 [pow(sums[j],
                                                                      self._input_state[
                                                                          j]) for j in
                                                                  considered_columns_indices],
                                                                 1)

        permanent *= pow(-1, sum(self._input_state))
        return permanent



__author__ = "Tomasz Rybotycki"

"""
    This class is meant to parallelize the CH Permanent Calculator.
"""

from multiprocessing.pool import Pool
from typing import Optional

from numpy import complex128, ndarray

from ..permanent_calculators.chin_huh_permanent_calculator import ChinHuhPermanentCalculator


class ParallelChinHuhPermanentCalculator(ChinHuhPermanentCalculator):

    def __init__(self, matrix: ndarray, input_state: Optional[ndarray] = None,
                 output_state: Optional[ndarray] = None) -> None:
        super().__init__(matrix, input_state, output_state)

    def calculate(self) -> complex128:
        """
            This is the main method of the calculator. Assuming that input state, output state and the matrix are
            defined correctly (that is we've got m x m matrix, and vectors of with length m) this calculates the
            permanent of an effective scattering matrix related to probability of obtaining output state from given
            input state.

            This method uses multiprocessing!

            :return: Permanent of effective scattering matrix.
        """
        if not self._can_calculation_be_performed():
            raise AttributeError

        v_vectors = self._calculate_v_vectors()

        permanent = complex128(0)

        with Pool() as p:
            results = p.map(self._compute_permanent_addend, v_vectors)
        permanent += sum(results)
        permanent /= pow(2, sum(self._input_state))

        return permanent



__author__ = "Tomasz Rybotycki"

"""
    The aim of this script is to provide a method returning the update indices for 
    Gray code iteration. The update indices makes sense in Gray code, as the only
    operation here is basically bit-flipping.
"""

from typing import List


def get_gray_code_update_indices(n: int) -> List[int]:
    """
        Computes and returns the update indices for Gray code iteration.
        :param n: The length of the code.
        :return: The list of indices to bit-flip during iteration.
    """
    if n <= 0:
        return []
    if n == 1:
        return [0]  # We index our tables from 0

    subproblem_update_indices = get_gray_code_update_indices(n - 1)
    update_indices = subproblem_update_indices.copy()
    update_indices.append(n - 1)
    update_indices.extend(subproblem_update_indices.__reversed__())
    return update_indices



__author__ = "Tomasz Rybotycki"

"""
    This script contains quick test ground of the generator.
"""

from ..src.GuanCodeGenerator import GuanCodeGenerator

for code in GuanCodeGenerator.generate_guan_codes([4, 0, 3, 3, 3]):
    print(code)



__author__ = "Tomasz Rybotycki"

"""
    This file holds the implementation of Chin and Huhs method of permanent calculation,
    as presented in [2].
    
    We use 
"""

import operator
from functools import reduce
from typing import Optional

from numpy import complex128, ndarray, nonzero, ones, zeros

from .bs_permanent_calculator_base import BSPermanentCalculatorBase


class ChinHuhPermanentCalculator(BSPermanentCalculatorBase):
    """
        This class is designed to calculate permanent of effective scattering matrix of
        a boson sampling instance. Note, that it can be used to calculate permanent of
        given matrix. All that is required that input and output
        states are set to [1, 1, ..., 1] with proper dimensions.
    """

    def __init__(self, matrix: ndarray, input_state: Optional[ndarray] = None,
                 output_state: Optional[ndarray] = None) -> None:
        super().__init__(matrix, input_state, output_state)

    def compute_permanent(self) -> complex128:
        """
            This is the main method of the calculator. Assuming that input state,
            output state and the matrix are defined correctly (that is we've got m x m
            matrix, and vectors of with length m) this calculates the permanent of an
            effective scattering matrix related to probability of obtaining output state
            from a given input state.

            :return: Permanent of effective scattering matrix.
        """
        if not self._can_calculation_be_performed():
            raise AttributeError

        # Required for Guan Code iteration
        v_vector = zeros(len(self._input_state), dtype=int)  # g
        code_update_information = ones(len(self._input_state), dtype=int)  # u
        position_limits = list(self._input_state)  # n

        permanent = complex128(0)
        sums = dict()

        binomials_product = 1
        considered_columns_indices = nonzero(self._output_state)[0]

        multiplier = 1

        # Initialization (0-th step).
        for i in considered_columns_indices:
            sums[i] = 0
            for j in range(len(self._input_state)):
                sums[i] += self._input_state[j] * self._matrix[i][j]

        permanent += multiplier * binomials_product * \
                     reduce(operator.mul, [pow(sums[i], self._output_state[i])
                                           for i in considered_columns_indices], 1)

        # Rest of the steps.
        while v_vector[-1] <= position_limits[-1]:

            # UPDATE R VECTOR
            index_to_update = 0  # i
            updated_value_at_index = v_vector[0] + code_update_information[0]  # k
            while updated_value_at_index > position_limits[index_to_update] \
                    or updated_value_at_index < 0:
                code_update_information[index_to_update] = -code_update_information[
                    index_to_update]
                index_to_update += 1

                if index_to_update == len(v_vector):

                    for _ in range(sum(self._input_state)):
                        permanent /= 2

                    return permanent

                updated_value_at_index = \
                    v_vector[index_to_update] + \
                    code_update_information[index_to_update]

            last_value_at_index = v_vector[index_to_update]
            v_vector[index_to_update] = updated_value_at_index
            # END UPDATE

            # START PERMANENT UPDATE
            multiplier = -multiplier

            # Sums update
            for i in sums:
                sums[i] -= 2 * (v_vector[index_to_update] - last_value_at_index) * \
                           self.matrix[i][index_to_update]

            # Binoms update
            if v_vector[index_to_update] > last_value_at_index:
                binomials_product *= (self._input_state[index_to_update] -
                                      last_value_at_index) / v_vector[index_to_update]
            else:
                binomials_product *= last_value_at_index / (
                        self._input_state[index_to_update] - v_vector[index_to_update])

            addend = multiplier * binomials_product * \
                     reduce(operator.mul, [pow(sums[j], self._output_state[j])
                                           for j in considered_columns_indices], 1)

            permanent += addend

        for _ in range(sum(self._input_state)):
            permanent /= 2

        return permanent



__author__ = "Tomasz Rybotycki"

"""
    This script contains implementation of a class that generates Guan codes (generalized Gray codes). I've changed the
    notation in comparison to 1998 Guan's work, but I've left the comments indicating which variable is which in the 
    original code. I've also modified the method slightly so that I can work out some python magic with reverse and also
    don't have additional array/list entries for control.
"""

from typing import List, Tuple

from numpy import ones, zeros

class GuanCodeGenerator:
    def __init__(self) -> None:
        pass

    @staticmethod
    def generate_guan_codes(positions_limits: List[int]) -> List[List[int]]:

        current_code = zeros(len(positions_limits), dtype=int)              # g
        code_update_information = ones(len(positions_limits), dtype=int)    # u
        position_limits = list(reversed(positions_limits))                  # n

        codes = []

        while(current_code[-1]<=position_limits[-1]):
            codes.append(list(reversed(current_code)))

            index_to_update = 0                                                         # i
            updated_value_at_index = current_code[0] + code_update_information[0]       # k
            while updated_value_at_index > position_limits[index_to_update] or updated_value_at_index < 0:
                code_update_information[index_to_update] = -code_update_information[index_to_update]
                index_to_update += 1

                if index_to_update == len(current_code):
                    return codes

                updated_value_at_index = current_code[index_to_update] + code_update_information[index_to_update]

            current_code[index_to_update] = updated_value_at_index

        return codes

    @staticmethod
    def generate_guan_codes_tuples(positions_limits: List[int]) -> List[Tuple[int]]:
        current_code = zeros(len(positions_limits), dtype=int)  # g
        code_update_information = ones(len(positions_limits), dtype=int)  # u
        position_limits = list(reversed(positions_limits))  # n

        codes = []

        while (current_code[-1] <= position_limits[-1]):
            codes.append(tuple(reversed(current_code)))

            index_to_update = 0  # i
            updated_value_at_index = current_code[0] + code_update_information[0]  # k
            while updated_value_at_index > position_limits[
                index_to_update] or updated_value_at_index < 0:
                code_update_information[index_to_update] = -code_update_information[
                    index_to_update]
                index_to_update += 1

                if index_to_update == len(current_code):
                    return codes

                updated_value_at_index = current_code[index_to_update] + \
                                         code_update_information[index_to_update]

            current_code[index_to_update] = updated_value_at_index

        return codes



__author__ = "Tomasz Rybotycki"

from copy import copy
from typing import List

from numpy import append, array, int64, ndarray

from ..boson_sampling_utilities.boson_sampling_utilities import particle_state_to_modes_state
from ..boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_interface import \
    BSPermanentCalculatorInterface
from .generalized_cliffords_simulation_strategy import GeneralizedCliffordsSimulationStrategy


class GeneralizedCliffordsSimulationStrategyInterfaceV2(GeneralizedCliffordsSimulationStrategy):
    def __init__(self, bs_permanent_calculator: BSPermanentCalculatorInterface) -> None:
        super().__init__(bs_permanent_calculator)

    def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:
        """
            Returns sample from linear optics experiments given output state.

            :param input_state: Input state in particle basis.
            :param samples_number: Number of samples to simulate.
            :return: A resultant state after traversing through interferometer.
        """
        self.input_state = input_state
        self.number_of_input_photons = sum(input_state)
        self._get_sorted_possible_states()
        self.pmfs = dict()

        samples = []

        while len(samples) < samples_number:
            self._fill_r_sample()
            samples.append(particle_state_to_modes_state(array(self.r_sample, dtype=int64), len(self.input_state)))
        return samples

    def _fill_r_sample(self) -> None:
        self.r_sample = []
        self.current_key = tuple(self.r_sample)
        self.current_sample_probability = 1

        while self.number_of_input_photons > len(self.r_sample):
            if self.current_key not in self.pmfs:
                self._calculate_new_layer_of_pmfs()
            self._sample_from_latest_pmf()

    def _calculate_new_layer_of_pmfs(self) -> None:
        number_of_particle_to_sample = len(self.r_sample) + 1
        possible_input_states = self._labeled_states[number_of_particle_to_sample]
        corresponding_k_vectors = [[self.input_state[i] - state[i] for i in range(len(state))]
                                   for state in possible_input_states]

        pmf = []

        weights = self._calculate_weights_from_k_vectors(array(corresponding_k_vectors, dtype=float))
        weights /= sum(weights)
        self.possible_outputs[self.current_key] = self._generate_possible_output_states()

        for output in self.possible_outputs[self.current_key]:
            output = particle_state_to_modes_state(output, len(self.input_state))
            pmf.append(0)
            for i in range(len(possible_input_states)):
                probability = self._calculate_outputs_probability(possible_input_states[i], output)
                probability *= weights[i]
                pmf[-1] += probability

        self.pmfs[self.current_key] = pmf

    def _generate_possible_output_states(self) -> List[ndarray]:
        possible_output_states = []

        for i in range(len(self.input_state)):
            new_possible_output = copy(self.r_sample)
            new_possible_output = append(new_possible_output, [i])
            possible_output_states.append(array(new_possible_output, dtype=int64))

        return possible_output_states



__author__ = "Tomasz Rybotycki"

"""
    This file holds a factory for simulation strategies.
"""

import enum
from copy import deepcopy

from .cliffords_r_simulation_strategy import CliffordsRSimulationStrategy
from .fixed_loss_simulation_strategy import FixedLossSimulationStrategy
from .generalized_cliffords_simulation_strategy import GeneralizedCliffordsSimulationStrategy
from .generalized_cliffords_uniform_losses_simulation_strategy import \
    GeneralizedCliffordsUniformLossesSimulationStrategy
from .lossy_networks_generalized_cliffords_simulation_strategy import \
    LossyNetworksGeneralizedCliffordsSimulationStrategy
from .lossy_state_approximated_simulation_strategy import LossyStateApproximationSimulationStrategy
from .nonuniform_losses_approximation_strategy import NonuniformLossesApproximationStrategy
from .simulation_strategy_interface import SimulationStrategyInterface
from .uniform_loss_simulation_strategy import UniformLossSimulationStrategy
from ..boson_sampling_utilities.permanent_calculators.bs_permanent_calculator_interface import \
    BSPermanentCalculatorInterface
from ..distribution_calculators.bs_distribution_calculator_interface import BosonSamplingExperimentConfiguration


class StrategyType(enum.IntEnum):
    FIXED_LOSS = enum.auto()
    UNIFORM_LOSS = enum.auto()
    CLIFFORD_R = enum.auto()
    GENERALIZED_CLIFFORD = enum.auto()
    LOSSY_NET_GENERALIZED_CLIFFORD = enum.auto()
    LOSSLESS_MODES_STRATEGY = enum.auto()
    GENERALIZED_U_LOSSY_CLIFFORD = enum.auto()
    NONUNIFORM_APPROXIMATED = enum.auto()
    UNIFORM_LOSSY_STATE_APPROXIMATED = enum.auto()


class SimulationStrategyFactory:
    def __init__(self, experiment_configuration: BosonSamplingExperimentConfiguration,
                 bs_permanent_calculator: BSPermanentCalculatorInterface,
                 strategy_type: StrategyType = StrategyType.FIXED_LOSS) -> None:
        self._experiment_configuration = experiment_configuration
        self._strategy_type = strategy_type
        self._bs_permanent_calculator = deepcopy(bs_permanent_calculator)
        self._strategy_mapping = {
            StrategyType.FIXED_LOSS: self._generate_fixed_losses_strategy,
            StrategyType.UNIFORM_LOSS: self._generate_uniform_losses_strategy,
            StrategyType.CLIFFORD_R: self._generate_r_cliffords_strategy,
            StrategyType.GENERALIZED_CLIFFORD: self._generate_generalized_cliffords_strategy,
            StrategyType.LOSSY_NET_GENERALIZED_CLIFFORD: self._generate_lossy_net_generalized_cliffords_strategy,
            StrategyType.GENERALIZED_U_LOSSY_CLIFFORD: self._generate_u_lossy_generalized_cliffords_strategy,
            StrategyType.NONUNIFORM_APPROXIMATED: self._generate_nonuniform_approximating_strategy,
            StrategyType.UNIFORM_LOSSY_STATE_APPROXIMATED: self._generate_uniform_lossy_state_approximating_strategy
        }

    @property
    def strategy_type(self) -> StrategyType:
        return self._strategy_type

    @strategy_type.setter
    def strategy_type(self, strategy_type: StrategyType) -> None:
        self._strategy_type = strategy_type

    @property
    def experiment_configuration(self) -> BosonSamplingExperimentConfiguration:
        return self._experiment_configuration

    @experiment_configuration.setter
    def experiment_configuration(self, experiment_configuration: BosonSamplingExperimentConfiguration) -> None:
        self._experiment_configuration = experiment_configuration

    @property
    def bs_permanent_calculator(self) -> BSPermanentCalculatorInterface:
        return self._bs_permanent_calculator

    @bs_permanent_calculator.setter
    def bs_permanent_calculator(self, bs_permanent_calculator: BSPermanentCalculatorInterface) -> None:
        self._bs_permanent_calculator = deepcopy(bs_permanent_calculator)

    def generate_strategy(self) -> SimulationStrategyInterface:
        """
            Generates simulation strategy of desired type. The type is selected in the constructor.
        :return: Simulation strategy of desired type.
        """
        handler = self._strategy_mapping.get(self._strategy_type, self._generate_fixed_losses_strategy)
        return handler()

    def _generate_fixed_losses_strategy(self) -> FixedLossSimulationStrategy:
        """
            Generates fixed loss strategy.
        :return: Fixed loss strategy.
        """
        return FixedLossSimulationStrategy(
            interferometer_matrix=self._experiment_configuration.interferometer_matrix,
            number_of_photons_left=self._experiment_configuration.number_of_particles_left,
            number_of_observed_modes=self._experiment_configuration.number_of_modes,
            network_simulation_strategy=self._experiment_configuration.network_simulation_strategy
        )

    def _generate_uniform_losses_strategy(self) -> UniformLossSimulationStrategy:
        """
            Generates uniform losses strategy.
        :return: Uniform losses strategy.
        """
        return UniformLossSimulationStrategy(
            self._experiment_configuration.interferometer_matrix,
            self._experiment_configuration.number_of_modes,
            self._experiment_configuration.uniform_transmissivity
        )

    def _generate_r_cliffords_strategy(self) -> CliffordsRSimulationStrategy:
        """
            Generates Cliffords algorithm strategy using their code implemented in R.
        :return: Cliffords strategy in R.
        """
        return CliffordsRSimulationStrategy(self._experiment_configuration.interferometer_matrix)

    def _generate_generalized_cliffords_strategy(self) -> GeneralizedCliffordsSimulationStrategy:
        """
            Generates generalized Cliffords strategy from Oszmaniec / Brod.
        :return: Generalized Cliffords strategy.
        """
        return GeneralizedCliffordsSimulationStrategy(deepcopy(self.bs_permanent_calculator))

    def _generate_lossy_net_generalized_cliffords_strategy(self) \
            -> LossyNetworksGeneralizedCliffordsSimulationStrategy:
        """
            Generates generalized Cliffords strategy for lossy networks from Oszmaniec / Brod 2020.
        :return: Generalized Cliffords strategy for lossy networks.
        """
        return LossyNetworksGeneralizedCliffordsSimulationStrategy(deepcopy(self.bs_permanent_calculator))

    def _generate_u_lossy_generalized_cliffords_strategy(self) -> GeneralizedCliffordsUniformLossesSimulationStrategy:
        return GeneralizedCliffordsUniformLossesSimulationStrategy(deepcopy(self.bs_permanent_calculator),
                                                                   self._experiment_configuration.uniform_transmissivity)

    def _generate_nonuniform_approximating_strategy(self) -> NonuniformLossesApproximationStrategy:
        return NonuniformLossesApproximationStrategy(deepcopy(self.bs_permanent_calculator),
                                                     self._experiment_configuration.number_of_modes - self._experiment_configuration.hierarchy_level,
                                                     self._experiment_configuration.uniform_transmissivity)

    def _generate_uniform_lossy_state_approximating_strategy(self) -> LossyStateApproximationSimulationStrategy:
        return LossyStateApproximationSimulationStrategy(
            bs_permanent_calculator=self._bs_permanent_calculator,
            uniform_transmissivity=self._experiment_configuration.uniform_transmissivity,
            hierarchy_level=self._experiment_configuration.hierarchy_level
        )



__author__ = "Tomasz Rybotycki"

"""
    This file holds an implementation of generic boson sampling experiment simulator.
"""

from typing import List

from numpy import ndarray

from .simulation_strategies.simulation_strategy_interface import SimulationStrategyInterface


class BosonSamplingSimulator:

    def __init__(self, simulation_strategy: SimulationStrategyInterface) -> None:
        self._simulation_strategy = simulation_strategy

    def get_classical_simulation_results(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:
        return self._simulation_strategy.simulate(input_state, samples_number)



__author__ = "Tomasz Rybotycki"

from typing import List, Iterable

from numpy import float64, ndarray, zeros

from ..boson_sampling_simulator import BosonSamplingSimulator
from ..boson_sampling_utilities.boson_sampling_utilities import generate_possible_outputs
from ..distribution_calculators.bs_distribution_calculator_interface import \
    BosonSamplingExperimentConfiguration, BSDistributionCalculatorInterface
from ..simulation_strategies.simulation_strategy_interface import SimulationStrategyInterface


class BSSampleBasedDistributionCalculator(BSDistributionCalculatorInterface):
    def __init__(self, experiment_configuration: BosonSamplingExperimentConfiguration,
                 strategy: SimulationStrategyInterface, samples_number=5000, outcomes=None) -> None:
        self._configuration = experiment_configuration
        self._strategy = strategy
        self._outcomes = outcomes
        self._samples_number = samples_number

    @property
    def configuration(self) -> BosonSamplingExperimentConfiguration:
        return self._configuration

    @configuration.setter
    def configuration(self, new_configuration: BosonSamplingExperimentConfiguration) -> None:
        self._configuration = new_configuration

    @property
    def strategy(self) -> SimulationStrategyInterface:
        return self._strategy

    @strategy.setter
    def strategy(self, new_strategy: SimulationStrategyInterface) -> None:
        self._strategy = new_strategy

    @property
    def outcomes(self) -> List[ndarray]:
        return self._outcomes

    @outcomes.setter
    def outcomes(self, new_outcomes: List[ndarray]) -> None:
        self._outcomes = new_outcomes

    @property
    def samples_number(self) -> int:
        return self._samples_number

    @samples_number.setter
    def samples_number(self, new_samples_number: int) -> None:
        self._samples_number = new_samples_number

    def calculate_distribution(self) -> List[float]:
        return self.calculate_approximate_distribution()

    def calculate_probabilities_of_outcomes(self, outcomes: Iterable[Iterable[int]]) -> List[float]:
        self._outcomes = outcomes
        return self.calculate_approximate_distribution()

    def calculate_approximate_distribution(self, samples_number: int = 5000) -> List[float]:
        """
            Prepares the approximate distribution using boson sampling simulation method described by
            Oszmaniec and Brod. Obviously higher number of samples will generate better approximation.
            :return: Approximate distribution as a list.
        """

        if self._outcomes is not None:
            possible_outcomes = self._outcomes
        else:
            possible_outcomes = generate_possible_outputs(self.configuration.number_of_particles_left,
                                                          self.configuration.number_of_modes)

        simulator = BosonSamplingSimulator(self._strategy)

        outcomes_probabilities = zeros(len(possible_outcomes), dtype=float64)

        samples = simulator.get_classical_simulation_results(self.configuration.initial_state, samples_number)
        for sample in samples:
            for j in range(len(possible_outcomes)):
                # Check if obtained result is one of possible outcomes.
                if all(sample == possible_outcomes[j]):  # Expect all elements of resultant list to be True.
                    outcomes_probabilities[j] += 1
                    break

        outcomes_probabilities = outcomes_probabilities / samples_number

        return list(outcomes_probabilities)

    def get_outcomes_in_proper_order(self) -> List[ndarray]:
        return self._outcomes



__author__ = "Tomasz Rybotycki"

"""
    One should note, that R is required to use this module, as original Clifford's
    program is written in R. On my Windows 10, I am using anaconda and I had to add
    R_HOME env variable and R_path\bin, R_path\bin\x64 to the path. 
    https://cran.r-project.org/web/packages/BosonSampling/index.html
"""

from numpy import arange, array, array_split, int64, ndarray

from .simulation_strategy_interface import SimulationStrategyInterface

try:
    from rpy2 import robjects
    from rpy2.robjects import packages

    from typing import List

    from theboss.boson_sampling_utilities.boson_sampling_utilities import \
        particle_state_to_modes_state


    class CliffordsRSimulationStrategy(SimulationStrategyInterface):
        def __init__(self, interferometer_matrix: ndarray) -> None:
            self.interferometer_matrix = interferometer_matrix

            required_packages = ('BosonSampling', 'Rcpp', 'RcppArmadillo')

            # Check if required R packages are installed. Inform if not.
            if not all(packages.isinstalled(package) for package in required_packages):
                print('Some packages are missing! Missing packages:')
                for package in required_packages:
                    if not packages.isinstalled(package):
                        print(package)

            boson_sampling_package = packages.importr('BosonSampling')
            self.cliffords_r_sampler = boson_sampling_package.bosonSampler

        def set_matrix(self, interferometer_matrix: ndarray) -> None:
            self.interferometer_matrix = interferometer_matrix

        @staticmethod
        def _numpy_array_to_r_matrix(numpy_array: ndarray) -> robjects.r.matrix:
            rows_number, columns_number = numpy_array.shape
            # Transposition is required as R inserts columns, not rows.
            r_values = robjects.ComplexVector(
                [val for val in numpy_array.transpose().reshape(numpy_array.size)])
            return robjects.r.matrix(r_values, nrow=rows_number, ncol=columns_number)

        def simulate(self, initial_state: ndarray, samples_number: int = 1) -> List[
            ndarray]:
            """
                Simulate BS experiment for given input.

                Note:   The results of Clifford & Clifford method are given in the first
                        quantization description (mode assignment)!

                :param initial_state:   Input state in the modes occupation description.
                :param samples_number:  Number of samples to sample.

                :return:    List of samples in the first quantization description (mode
                            assignment)
            """
            number_of_bosons = int(sum(initial_state))

            boson_sampler_input_matrix = self._numpy_array_to_r_matrix(
                self.interferometer_matrix[:, arange(number_of_bosons)])

            result, permanent, probability_mass_function = \
                self.cliffords_r_sampler(boson_sampler_input_matrix,
                                         sampleSize=samples_number,
                                         perm=False)

            # Add -1 to R indexation of modes (they start from 1).
            python_result = array([mode_value - 1 for mode_value in result],
                                  dtype=int64)
            samples_in_particle_states = array_split(python_result, samples_number)

            # There are some problems with the actual and theoretical runtimes. The
            # reason for that could be parsing the result to a second quantization
            # description.
            # return samples_in_particle_states
            samples_in_occupation_description = []

            for sample in samples_in_particle_states:
                samples_in_occupation_description.append(
                    particle_state_to_modes_state(sample))

            return samples_in_occupation_description




except ImportError as e:
    print("You have to install rpy2 package in order to use this strategy.")


    class CliffordsRSimulationStrategy(SimulationStrategyInterface):
        def __init__(self, interferometer_matrix: ndarray) -> None:
            raise NotImplementedError

        def simulate(self, input_state: ndarray, samples_number: int = 1) -> List[ndarray]:
            raise NotImplementedError



__author__ = "Tomasz Rybotycki"

"""
    This file holds an interface for Boson Sampling Permanent Calculators. Boson Sampling (BS) Permanent calculation is
    different in that way, that it requires computing (or rather BS is described by) effective scattering matrix thus
    the permanent returned is not necessarily the permanent of a specified matrix. In order to compute the permanent
    of a given matrix one should specify input and output states as [1, ... , 1]. 
"""

import abc

from numpy import complex128, ndarray


class BSPermanentCalculatorInterface(abc.ABC):

    @abc.abstractmethod
    def compute_permanent(self) -> complex128:
        """Computes permanent of a matrix given before."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def matrix(self) -> ndarray:
        raise NotImplementedError

    @matrix.setter
    @abc.abstractmethod
    def matrix(self, matrix: ndarray) -> None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def input_state(self) -> ndarray:
        raise NotImplementedError

    @input_state.setter
    @abc.abstractmethod
    def input_state(self, input_state: ndarray) -> None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def output_state(self) -> ndarray:
        raise NotImplementedError

    @output_state.setter
    @abc.abstractmethod
    def output_state(self, output_state: ndarray) -> None:
        raise NotImplementedError



__author__ = "Tomasz Rybotycki"

from copy import deepcopy
from typing import List, Iterable

from numpy import ndarray
from scipy import special

from ..boson_sampling_utilities.boson_sampling_utilities import generate_possible_outputs
from ..distribution_calculators.bs_distribution_calculator_interface import \
    BosonSamplingExperimentConfiguration
from ..distribution_calculators.bs_distribution_calculator_with_fixed_losses import \
    BSDistributionCalculatorWithFixedLosses, BSPermanentCalculatorInterface
from multiprocessing import cpu_count, Pool


class BSDistributionCalculatorWithUniformLosses \
            (BSDistributionCalculatorWithFixedLosses):
    def __init__(self, configuration: BosonSamplingExperimentConfiguration,
                 permanent_calculator: BSPermanentCalculatorInterface) -> None:
        super().__init__(configuration, permanent_calculator)
        self.weights = self._initialize_weights()
        self.weightless = False

    def set_weightless(self, weightless: bool) -> None:
        if not weightless:
            self.weights = self._initialize_weights()
        else:
            self.weights = [1 for _ in self.weights]

        self.weightless = weightless

    def _initialize_weights(self) -> List[float]:

        weight = \
            lambda n, l, eta: pow(eta, l) * special.binom(n, l) * pow(1.0 - eta, n - l)
        weights = []

        for number_of_particles_left \
                in range(self.configuration.initial_number_of_particles + 1):
            weights.append(
                weight(self.configuration.initial_number_of_particles,
                       number_of_particles_left,
                       self.configuration.uniform_transmissivity)
            )

        return weights

    def calculate_probabilities_of_outcomes(self,
                                            outcomes: Iterable[Iterable[int]]) -> \
    List[float]:

        with Pool(processes=cpu_count()) as pool:
            outcomes_probabilities = pool.map(self._calculate_probability_of_outcome,
                                              outcomes)

        return outcomes_probabilities

    def _calculate_probability_of_outcome(self, outcome: ndarray) -> float:

        number_of_particles_left = int(sum(outcome))
        l = number_of_particles_left

        if l == 0:
            return self.weights[0]

        n = self.configuration.initial_number_of_particles

        subconfiguration = deepcopy(self.configuration)

        subconfiguration.number_of_particles_left = number_of_particles_left
        subconfiguration.number_of_particles_lost = n - l
        subdistribution_calculator = \
            BSDistributionCalculatorWithFixedLosses(subconfiguration,
                                                    self._permanent_calculator)

        probability_of_outcome = subdistribution_calculator.calculate_probabilities_of_outcomes([outcome])[0]

        return probability_of_outcome * self.weights[l]


    def get_outcomes_in_proper_order(self) -> List[ndarray]:
        return generate_possible_outputs(self.configuration.initial_number_of_particles,
                                         self.configuration.number_of_modes, consider_loses=True)



__author__ = "Tomasz Rybotycki"

"""
    This script contains the implementation of submatrices calculation method as used
    in Clifford & Clifford version B of the algorithm. C&C algorithm uses Laplace
    expansion for the permanents in order to compute the set of probabilities in each
    step of the algorithm. Instead of computing each permanent separately, we can
    compute them all in one run, which is vastly more efficient.
    
    Instead of using Rysers formula, or Glynn formula, we base our approach on the
    Chin-Huh's formula, which takes into account possible bunching in the input states
    (or rather can be interpreted as that). 
"""

import abc
from numpy import complex128, ndarray, int64, array, nonzero, zeros, ones
import operator
from functools import reduce
from typing import List, Optional


class BSSubmatricesPermanentCalculatorInterface(abc.ABC):
    """
        This is the interface class for submatrices permanents calculator. For now
        BSCCCHSubmatricesPermanentCalculator is the only class of that kind, but it's
        possible that there will be more (based e.g. on Glynns formula that grants
        numerical stability.

        If the above happens, place the interface class in the separate file.
    """

    @abc.abstractmethod
    def compute_permanents(self) -> List[complex128]:
        """Computes permanent of a matrix given before."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def matrix(self) -> ndarray:
        raise NotImplementedError

    @matrix.setter
    @abc.abstractmethod
    def matrix(self, matrix: ndarray) -> None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def input_state(self) -> ndarray:
        raise NotImplementedError

    @input_state.setter
    @abc.abstractmethod
    def input_state(self, input_state: ndarray) -> None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def output_state(self) -> ndarray:
        raise NotImplementedError

    @output_state.setter
    @abc.abstractmethod
    def output_state(self, output_state: ndarray) -> None:
        raise NotImplementedError


class BSSubmatricesPermanentCalculatorBase(BSSubmatricesPermanentCalculatorInterface,
                                           abc.ABC):
    """
        Base class for BSSubmatricesPermanentCalculator classes. It takes care of some
        boilerplate code.

        Again, it should be put into separate file were the
        BSCCCHSubmatricesPermanentCalculator cease to be the only submatrices
        permanent calculator.
    """

    def __init__(self, matrix: ndarray, input_state: Optional[ndarray] = None,
                 output_state: Optional[ndarray] = None) -> None:
        if output_state is None:
            output_state = array([], dtype=int64)
        if input_state is None:
            input_state = array([], dtype=int64)
        self._matrix = matrix
        self._input_state = input_state
        self._output_state = output_state

    @property
    def matrix(self) -> ndarray:
        return self._matrix

    @matrix.setter
    def matrix(self, matrix: ndarray) -> None:
        self._matrix = matrix

    @property
    def input_state(self) -> ndarray:
        return self._input_state

    @input_state.setter
    def input_state(self, input_state: ndarray) -> None:
        self._input_state = array(input_state, dtype=int64)

    @property
    def output_state(self) -> ndarray:
        return self._output_state

    @output_state.setter
    def output_state(self, output_state: ndarray) -> None:
        self._output_state = array(output_state, dtype=int64)


class BSCCCHSubmatricesPermanentCalculator(BSSubmatricesPermanentCalculatorBase):

    """
        The name stands for Boson Sampling Clifford & Clifford Chin-Huh submatrices
        permanent calculator, as it uses Clifford & Clifford approach to compute
        permanents of submatrices to compute sub-distribution of Boson Sampling problem
        instance. The starting point in our case is Chin-Huh permanent calculator
        iterated in Guan Codes induced order.
    """

    def compute_permanents(self) -> List[complex128]:

        # TODO TR:  This method is huge and complicated. It would be smart to break
        #           it down into smaller ones.

        permanents = [complex128(0) for _ in range(len(self.input_state))]

        # Required for Guan Code iteration
        v_vector = zeros(len(self._input_state), dtype=int)  # g
        code_update_information = ones(len(self._input_state), dtype=int)  # u
        position_limits = list(self._input_state)  # n

        sums = dict()

        binomials_product = 1
        considered_columns_indices = nonzero(self._output_state)[0]

        multiplier = 1

        # Initialization (0-th step).
        for i in considered_columns_indices:
            sums[i] = 0
            for j in range(len(self._input_state)):
                sums[i] += self._input_state[j] * self._matrix[i][j]

        for i in range(len(permanents)):
            if v_vector[i] != self.input_state[i]:
                permanents[i] += multiplier * binomials_product * \
                             reduce(operator.mul, [pow(sums[i], self._output_state[i])
                                                   for i in considered_columns_indices], 1)

        # Rest of the steps.
        while v_vector[-1] <= position_limits[-1]:

            # UPDATE R VECTOR
            index_to_update = 0  # i
            updated_value_at_index = v_vector[0] + code_update_information[0]  # k
            while updated_value_at_index > position_limits[index_to_update] \
                    or updated_value_at_index < 0:
                code_update_information[index_to_update] = -code_update_information[
                    index_to_update]
                index_to_update += 1

                if index_to_update == len(v_vector):

                    for _ in range(sum(self.input_state) - 1):
                        for i in range(len(permanents)):
                            permanents[i] /= 2

                    return permanents

                updated_value_at_index = \
                    v_vector[index_to_update] + \
                    code_update_information[index_to_update]

            last_value_at_index = v_vector[index_to_update]
            v_vector[index_to_update] = updated_value_at_index
            # END UPDATE

            # START PERMANENT UPDATE
            multiplier = -multiplier

            # Sums update
            for i in sums:
                sums[i] -= 2 * (v_vector[index_to_update] - last_value_at_index) * \
                           self.matrix[i][index_to_update]

            # Binoms update
            if v_vector[index_to_update] > last_value_at_index:
                binomials_product *= (self._input_state[index_to_update] -
                                      last_value_at_index) / v_vector[index_to_update]
            else:
                binomials_product *= last_value_at_index / (
                        self._input_state[index_to_update] - v_vector[index_to_update])

            addend = multiplier * binomials_product * \
                     reduce(operator.mul, [pow(sums[j], self._output_state[j])
                                           for j in considered_columns_indices], 1)

            for i in range(len(permanents)):
                if v_vector[i] != self.input_state[i]:
                    permanents[i] += addend

        for _ in range(sum(self._input_state) - 1):
            for i in range(len(permanents)):
                permanents[i] /= 2

        return permanents
